import pytest
import sys
import textwrap
import rpy2.robjects as robjects
import rpy2.robjects.methods as methods
rinterface = robjects.rinterface


def test_set_accessors():
    robjects.r['setClass']("A", robjects.r('list(foo="numeric")'))
    robjects.r['setMethod']("length", signature="A",
                            definition = robjects.r("function(x) 123"))
    class A(methods.RS4):
        def __init__(self):
            obj = robjects.r['new']('A')
            super().__init__(obj)

    acs = (('length', None, True, None), )
    methods.set_accessors(A, "A", None, acs)
    a = A()
    assert a.length[0] == 123


def test_RS4Type_noaccessors():
    robjects.r['setClass']("Foo", robjects.r('list(foo="numeric")'))
    classdef = """
    from rpy2 import robjects
    from rpy2.robjects import methods
    class Foo(methods.RS4, metaclass=methods.RS4_Type):
        def __init__(self):
            obj = robjects.r['new']('Foo')
            super().__init__(obj)
    """
    code = compile(textwrap.dedent(classdef),
                   '<string>', 'exec')
    ns = dict()
    exec(code, ns)
    f = ns['Foo']()
    # TODO: test ?


def test_RS4_factory():
    rclassname = 'Foo'
    robjects.r['setClass'](rclassname, 
                           robjects.r('list(bar="numeric")'))
    obj = robjects.r['new'](rclassname)
    f_rs4i = methods.rs4instance_factory(obj)
    assert rclassname == type(f_rs4i).__name__


def test_RS4Type_accessors():
    robjects.r['setClass']("R_A", robjects.r('list(foo="numeric")'))
    robjects.r['setMethod']("length", signature="R_A",
                            definition = robjects.r("function(x) 123"))

    classdef = """
    from rpy2 import robjects
    from rpy2.robjects import methods
    class R_A(methods.RS4, metaclass=methods.RS4_Type):
        __accessors__ = (
            ('length', None,
             'get_length', False, 'get the length'),
            ('length', None,
             None, True, 'length'))
        def __init__(self):
            obj = robjects.r['new']('R_A')
            super().__init__(obj)            
    """
    code = compile(textwrap.dedent(classdef),
                   '<string>', 'exec')
    ns = dict()
    exec(code, ns)
    R_A = ns['R_A']
    class A(R_A):
        __rname__ = 'R_A'

    ra = R_A()
    assert ra.get_length()[0] == 123
    assert ra.length[0] == 123

    a = A()
    assert a.get_length()[0] == 123
    assert a.length[0] == 123


def test_getclassdef():
    robjects.r('library(stats4)')
    cr = methods.getclassdef('mle', 'stats4')
    assert not cr.virtual


def test_RS4Auto_Type():
    robjects.r('library(stats4)')
    class MLE(object,
              metaclass=robjects.methods.RS4Auto_Type):
        __rname__ = 'mle'
        __rpackagename__ = 'stats4'
    # TODO: test ?


def test_RS4Auto_Type_nopackname():
    robjects.r('library(stats4)')

    class MLE(object,
              metaclass=robjects.methods.RS4Auto_Type):
        __rname__ = 'mle'
    # TODO: test ?



import pytest
import rpy2.rlike.container as rlc


class TestOrdDict(object):

    def test_new(self):
        nl = rlc.OrdDict()

        x = (('a', 123), ('b', 456), ('c', 789))
        nl = rlc.OrdDict(x)

    def test_new_invalid(self):
        with pytest.raises(TypeError):
            rlc.OrdDict({})

    @pytest.mark.parametrize('methodname,args',
                             (('__cmp__', [None]),
                              ('__eq__', [None]),
                              ('__ne__', [None]),
                              ('__reversed__', []),
                              ('sort', [])))
    def test_notimplemented(self, methodname, args):
        nl = rlc.OrdDict()
        with pytest.raises(NotImplementedError):
            getattr(nl, methodname)(*args)

    def test_repr(self):
        x = (('a', 123), ('b', 456), ('c', 789))
        nl = rlc.OrdDict(x)
        assert isinstance(repr(nl), str)

    def test_iter(self):
        x = (('a', 123), ('b', 456), ('c', 789))
        nl = rlc.OrdDict(x)
        for a, b in zip(nl, x):
            assert a == b[0]

    def test_len(self):
        x = rlc.OrdDict()
        assert len(x) == 0

        x['a'] = 2
        x['b'] = 1

        assert len(x) == 2

    def test_getsetitem(self):
        x = rlc.OrdDict()
        
        x['a'] = 1
        assert len(x) == 1
        assert x['a'] == 1
        assert x.index('a') == 0
        x['a'] = 2
        assert len(x) == 1
        assert x['a'] == 2
        assert x.index('a') == 0
        x['b'] = 1
        assert len(x) == 2
        assert x['b'] == 1
        assert x.index('b') == 1

    def test_get(self):
        x = rlc.OrdDict()
        x['a'] = 1
        assert x.get('a') == 1
        assert x.get('b') is None
        assert x.get('b', 2) == 2
        
    def test_keys(self):
        x = rlc.OrdDict()
        word = 'abcdef'
        for i,k in enumerate(word):
            x[k] = i
        for i,k in enumerate(x.keys()):
            assert word[i] == k

    def test_getsetitemwithnone(self):
        x = rlc.OrdDict()
        
        x['a'] = 1
        x[None] = 2
        assert len(x) == 2
        x['b'] = 5
        assert len(x) == 3
        assert x['a'] == 1
        assert x['b'] == 5
        assert x.index('a') == 0
        assert x.index('b') == 2
        
    def test_reverse(self):
        x = rlc.OrdDict()
        x['a'] = 3
        x['b'] = 2
        x['c'] = 1
        x.reverse()
        assert x['c'] == 1
        assert x.index('c') == 0
        assert x['b'] == 2
        assert x.index('b') == 1
        assert x['a'] == 3
        assert x.index('a') == 2

    def test_items(self):
        args = (('a', 5), ('b', 4), ('c', 3),
                ('d', 2), ('e', 1))
        x = rlc.OrdDict(args)
        it = x.items()
        for ki, ko in zip(args, it):
            assert ki[0] ==  ko[0]
            assert ki[1] == ko[1]


class TestTaggedList(object):

    def test__add__(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        tl = tl + tl
        assert len(tl) == 6
        assert tl.tags == ('a', 'b', 'c', 'a', 'b', 'c')
        assert tuple(tl) == (1,2,3,1,2,3)

    def test__delitem__(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        assert len(tl) == 3
        del tl[1]
        assert len(tl) == 2
        assert tl.tags == ('a', 'c')
        assert tuple(tl) == (1, 3)

    def test__delslice__(self):
        tl = rlc.TaggedList((1,2,3,4), tags=('a', 'b', 'c', 'd'))        
        del tl[1:3]
        assert len(tl) == 2
        assert tl.tags == ('a', 'd')
        assert tuple(tl) == (1, 4)

    def test__iadd__(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        tl += tl
        assert len(tl) == 6
        assert tl.tags == ('a', 'b', 'c', 'a', 'b', 'c')
        assert tuple(tl) == (1,2,3,1,2,3)

    def test__imul__(self):
        tl = rlc.TaggedList((1,2), tags=('a', 'b'))
        tl *= 3
        assert len(tl) == 6
        assert tl.tags == ('a', 'b', 'a', 'b', 'a', 'b')
        assert tuple(tl) == (1,2,1,2,1,2)

    def test__init__(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        with pytest.raises(ValueError):
            rlc.TaggedList((1,2,3), tags = ('b', 'c'))

    def test__setslice__(self):
        tl = rlc.TaggedList((1,2,3,4), tags=('a', 'b', 'c', 'd'))        
        tl[1:3] = [5, 6]
        assert len(tl) == 4
        assert tl.tags == ('a', 'b', 'c', 'd')
        assert tuple(tl) == (1, 5, 6, 4)

    def test_append(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        assert len(tl) == 3
        tl.append(4, tag='a')
        assert len(tl) == 4
        assert tl[3] == 4
        assert tl.tags == ('a', 'b', 'c', 'a')

    def test_extend(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        tl.extend([4, 5])
        assert tuple(tl.itertags()) == ('a', 'b', 'c', None, None)
        assert tuple(tl) == (1, 2, 3, 4, 5)

    def test_insert(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        tl.insert(1, 4, tag = 'd')
        assert tuple(tl.itertags()) == ('a', 'd', 'b', 'c')
        assert tuple(tl) == (1, 4, 2, 3)

    def test_items(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))        
        assert tuple(tl.items()) == (('a', 1), ('b', 2), ('c', 3))

    def test_iterontag(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'a'))
        assert tuple(tl.iterontag('a')) == (1, 3)

    def test_itertags(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))
        assert tuple(tl.itertags()) == ('a', 'b', 'c')

    def test_pop(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))
        assert len(tl) == 3
        elt = tl.pop()
        assert elt == 3
        assert len(tl) == 2
        assert tl.tags == ('a', 'b')
        assert tuple(tl) == (1, 2)

        elt = tl.pop(0)
        assert elt == 1
        assert len(tl) == 1
        assert tl.tags == ('b', )

    def test_remove(self):
        tl = rlc.TaggedList((1,2,3), tags=('a', 'b', 'c'))
        assert len(tl) == 3
        tl.remove(2)
        assert len(tl) == 2
        assert tl.tags == ('a', 'c')
        assert tuple(tl) == (1, 3)

    def test_reverse(self):
        tn = ['a', 'b', 'c']
        tv = [1,2,3]
        tl = rlc.TaggedList(tv, tags = tn)
        tl.reverse()
        assert len(tl) == 3
        assert tl.tags == ('c', 'b', 'a')
        assert tuple(tl) == (3, 2, 1)

    def test_sort(self):
        tn = ['a', 'c', 'b']
        tv = [1,3,2]
        tl = rlc.TaggedList(tv, tags = tn)
        tl.sort()

        assert tl.tags == ('a', 'b', 'c')
        assert tuple(tl) == (1, 2, 3)

    def test_tags(self):
        tn = ['a', 'b', 'c']
        tv = [1,2,3]
        tl = rlc.TaggedList(tv, tags = tn)
        tags = tl.tags
        assert isinstance(tags, tuple) is True
        assert tags == ('a', 'b', 'c')

        tn = ['d', 'e', 'f']
        tl.tags = tn
        assert isinstance(tags, tuple) is True
        assert tl.tags == tuple(tn)

    def test_settag(self):
        tn = ['a', 'b', 'c']
        tv = [1,2,3]
        tl = rlc.TaggedList(tv, tags = tn)
        tl.settag(1, 'z')
        assert tl.tags == ('a', 'z', 'c')

    def test_from_items(self):
        od = rlc.OrdDict( (('a', 1), ('b', 2), ('c', 3)) )
        tl = rlc.TaggedList.from_items(od)
        assert tl.tags == ('a', 'b', 'c')
        assert tuple(tl) == (1, 2, 3)

        tl = rlc.TaggedList.from_items({'a':1, 'b':2, 'c':3})
        assert set(tl.tags) == set(('a', 'b', 'c'))
        assert set(tuple(tl)) == set((1, 2, 3))



import enum
import os


class InterfaceType(enum.Enum):
    ABI = 1
    API = 2


class SignatureDefinition(object):

    def __init__(self, name, rtype, arguments):
        self.name = name
        self.rtype = rtype
        self.arguments = arguments

    @property
    def callback_def(self):
        return '{} ({})'.format(self.rtype, ' ,'.join(self.arguments))

    @property
    def extern_def(self):
        return '{} {}({})'.format(
            self.rtype, self.name, ' ,'.join(self.arguments)
        )

    @property
    def extern_python_def(self):
        return 'extern "Python" {} {}({});'.format(
            self.rtype, self.name, ' ,'.join(self.arguments)
        )


def get_ffi_mode(_rinterface_cffi):
    if hasattr(_rinterface_cffi, 'lib'):
        return InterfaceType.API
    else:
        return InterfaceType.ABI


def callback(definition, _rinterface_cffi):
    def decorator(func):
        if get_ffi_mode(_rinterface_cffi) == InterfaceType.ABI:
            res = _rinterface_cffi.ffi.callback(definition.callback_def)(func)
        elif get_ffi_mode(_rinterface_cffi) == InterfaceType.API:
            res = _rinterface_cffi.ffi.def_extern()(func)
        else:
            raise RuntimeError('The cffi mode is neither ABI or API.')
        return res
    return decorator


# Callbacks
_capsule_finalizer_def = SignatureDefinition('_capsule_finalizer',
                                             'void', ('SEXP',))
_evaluate_in_r_def = SignatureDefinition('_evaluate_in_r',
                                         'SEXP', ('SEXP args',))


_consoleflush_def = SignatureDefinition('_consoleflush', 'void', ('void', ))
if os.name == 'nt':
    _consoleread_def = SignatureDefinition('_consoleread', 'int',
                                           ('char *', 'char *',
                                            'int', 'int'))
else:
    _consoleread_def = SignatureDefinition('_consoleread', 'int',
                                           ('char *', 'unsigned char *',
                                            'int', 'int'))
_consolereset_def = SignatureDefinition('_consolereset', 'void', ('void', ))
_consolewrite_def = SignatureDefinition('_consolewrite', 'void',
                                        ('char *', 'int'))
_consolewrite_ex_def = SignatureDefinition('_consolewrite_ex', 'void',
                                           ('char *', 'int', 'int'))
_showmessage_def = SignatureDefinition('_showmessage', 'void', ('char *', ))
_choosefile_def = SignatureDefinition('_choosefile', 'int',
                                      ('int', 'char *', 'int'))
_cleanup_def = SignatureDefinition('_cleanup', 'void',
                                   ('SA_TYPE', 'int', 'int'))
_showfiles_def = SignatureDefinition('_showfiles', 'int',
                                     ('int', 'const char **', 'const char **',
                                      'const char *', 'Rboolean',
                                      'const char *'))
_processevents_def = SignatureDefinition('_processevents', 'void', ('void', ))
_busy_def = SignatureDefinition('_busy', 'void', ('int', ))
_callback_def = SignatureDefinition('_callback', 'void', ('void', ))
# TODO: should be const char *
_yesnocancel_def = SignatureDefinition('_yesnocancel', 'int', ('char *', ))

_parsevector_wrap_def = SignatureDefinition('_parsevector_wrap',
                                            'SEXP', ('void *data', ))

_handler_def = SignatureDefinition('_handler_wrap',
                                   'SEXP', ('SEXP cond', 'void *hdata'))

_exec_findvar_in_frame_def = SignatureDefinition('_exec_findvar_in_frame',
                                                 'void', ('void *data', ))



import atexit
import contextlib
import os
import math
import platform
import signal
import textwrap
import threading
import typing
import warnings
from typing import Union
from rpy2.rinterface_lib import openrlib
import rpy2.rinterface_lib._rinterface_capi as _rinterface
import rpy2.rinterface_lib.embedded as embedded
import rpy2.rinterface_lib.conversion as conversion
from rpy2.rinterface_lib.conversion import _cdata_res_to_rinterface
import rpy2.rinterface_lib.memorymanagement as memorymanagement
from rpy2.rinterface_lib import na_values
from rpy2.rinterface_lib.sexp import NULLType
import rpy2.rinterface_lib.bufferprotocol as bufferprotocol
from rpy2.rinterface_lib import sexp
from rpy2.rinterface_lib.sexp import CharSexp  # noqa: F401
from rpy2.rinterface_lib.sexp import RTYPES
from rpy2.rinterface_lib.sexp import SexpVector
from rpy2.rinterface_lib.sexp import StrSexpVector
from rpy2.rinterface_lib.sexp import Sexp
from rpy2.rinterface_lib.sexp import SexpEnvironment
from rpy2.rinterface_lib.sexp import unserialize  # noqa: F401
from rpy2.rinterface_lib.sexp import emptyenv
from rpy2.rinterface_lib.sexp import baseenv
from rpy2.rinterface_lib.sexp import globalenv

if os.name == 'nt':
    import rpy2.rinterface_lib.embedded_mswin as embedded_mswin
    embedded._initr = embedded_mswin._initr_win32

R_NilValue = openrlib.rlib.R_NilValue

endr = embedded.endr

_evaluation_context = globalenv


def get_evaluation_context() -> SexpEnvironment:
    """Get the frame (environment) in which R code is currently evaluated."""
    return _evaluation_context


@contextlib.contextmanager
def local_context(
        env: typing.Optional[SexpEnvironment] = None,
        use_rlock: bool = True
) -> typing.Iterator[SexpEnvironment]:
    """Local context for the evaluation of R code.

    Args:
    - env: an environment to use as a context. If not specified (None, the
      default), a child environment to the current context is created.
    - use_rlock: whether to use a threading lock (see the documentation about
      "rlock". The default is True.

    Returns:
    Yield the environment (passed to env, or created).
    """

    global _evaluation_context

    parent_frame = _evaluation_context
    if env is None:
        env = baseenv['new.env'](
            baseenv['parent.frame']()
            if parent_frame is None
            else parent_frame)
    try:
        if use_rlock:
            with openrlib.rlock:
                _evaluation_context = env
                yield env
        else:
            _evaluation_context = env
            yield env
    finally:
        _evaluation_context = parent_frame


def _sigint_handler(sig, frame):
    raise KeyboardInterrupt()


@_cdata_res_to_rinterface
def parse(text: str, num: int = -1):
    """Parse a string as R code.

    :param:`text` A string with R code to parse.
    :param:`num` The maximum number of lines to parse. If -1, no
      limit is applied.
    """

    if not isinstance(text, str):
        raise TypeError('text must be a string.')
    robj = StrSexpVector([text])
    with memorymanagement.rmemory() as rmemory:
        res = _rinterface._parse(robj.__sexp__._cdata, num, rmemory)
    return res


def evalr(source: str, maxlines: int = -1) -> sexp.Sexp:
    """Evaluate a string as R code.

    Evaluate a string as R just as it would happen when writing
    code in an R terminal.

    :param:`text` A string to be evaluated as R code.
    :param:`maxlines` The maximum number of lines to parse. If -1, no
      limit is applied."""

    res = parse(source, num=maxlines)
    res = baseenv['eval'](res)
    return res


def vector_memoryview(obj: sexp.SexpVector,
                      sizeof_str: str, cast_str: str) -> memoryview:
    """
    :param:`obj` R vector
    :param:`sizeof_str` Type in a string to use with ffi.sizeof()
        (for example "int")
    :param:`cast_str` Type in a string to use with memoryview.cast()
        (for example "i")
    """
    b = openrlib.ffi.buffer(
        obj._R_GET_PTR(obj.__sexp__._cdata),
        openrlib.ffi.sizeof(sizeof_str) * len(obj))
    shape = bufferprotocol.getshape(obj.__sexp__._cdata)
    # One could have expected to only need builtin Python
    # and do something like
    # ```
    # mv = memoryview(b).cast(cast_str, shape, order='F')
    # ```
    # but Python does not handle FORTRAN-ordered arrays without having
    # to write C extensions. We have to use numpy.
    # TODO: Having numpy a requirement just for this is a problem.
    # TODO: numpy needed for memoryview
    #   (as long as https://bugs.python.org/issue34778 not resolved)
    import numpy
    a = numpy.frombuffer(b, dtype=cast_str).reshape(shape, order='F')
    mv = memoryview(a)
    return mv


class SexpSymbol(sexp.Sexp):
    """An unevaluated R symbol."""

    def __init__(
            self,
            obj: Union[Sexp,
                       _rinterface.SexpCapsule,
                       _rinterface.UninitializedRCapsule,
                       str]
    ):
        if isinstance(obj, Sexp) or isinstance(obj, _rinterface.CapsuleBase):
            super().__init__(obj)
        elif isinstance(obj, str):
            name_cdata = _rinterface.ffi.new('char []', obj.encode('utf-8'))
            sexp = _rinterface.SexpCapsule(
                openrlib.rlib.Rf_install(name_cdata))
            super().__init__(sexp)
        else:
            raise TypeError(
                'The constructor must be called '
                'with that is an instance of rpy2.rinterface.sexp.Sexp '
                'or an instance of rpy2.rinterface._rinterface.SexpCapsule')

    def __str__(self) -> str:
        return conversion._cchar_to_str(
            openrlib._STRING_VALUE(
                self._sexpobject._cdata
            ), 'utf-8'
        )


class _MissingArgType(SexpSymbol, metaclass=sexp.SingletonABC):

    def __init__(self):
        if embedded.isready():
            tmp = sexp.Sexp(
                _rinterface.UnmanagedSexpCapsule(
                    openrlib.rlib.R_MissingArg
                )
            )
        else:
            tmp = sexp.Sexp(
                _rinterface.UninitializedRCapsule(RTYPES.SYMSXP.value)
            )
        super().__init__(tmp)

    def __bool__(self) -> bool:
        """This is always False."""
        return False

    @property
    def __sexp__(self) -> _rinterface.SexpCapsule:
        return self._sexpobject


MissingArg = _MissingArgType()


class SexpPromise(Sexp):

    @_cdata_res_to_rinterface
    def eval(self, env: typing.Optional[SexpEnvironment] = None) -> sexp.Sexp:
        """"Evalute the R "promise".

        :param:`env` The environment in which to evaluate the
          promise.
        """
        if env is None:
            env = globalenv
        return openrlib.rlib.Rf_eval(self.__sexp__._cdata, env)


NPCOMPAT_TYPE = typing.TypeVar('NPCOMPAT_TYPE',
                               'ByteSexpVector',
                               'BoolSexpVector',
                               'IntSexpVector',
                               'FloatSexpVector')


class NumpyArrayMixin:
    """Numpy-specific API for accessing the content of a numpy array.

    This interface implements version 3 of Numpy's `__array_interface__`
    and is only available / possible for some of the R vectors."""

    @property
    def __array_interface__(
            self: NPCOMPAT_TYPE
    ) -> dict:
        """Return an `__array_interface__` version 3.

        Note that the pointer returned in the items 'data' corresponds to
        a memory area under R's memory management and that it will become
        invalid once the area once R frees the object. It is safer to keep
        the rpy2 object proxying the R object alive for the duration the
        pointer is used in Python / numpy."""
        shape = bufferprotocol.getshape(self.__sexp__._cdata)
        data = openrlib.ffi.buffer(self._R_GET_PTR(self.__sexp__._cdata))
        strides = bufferprotocol.getstrides(self.__sexp__._cdata,
                                            shape,
                                            self._R_SIZEOF_ELT)
        return {'shape': shape,
                'typestr': self._NP_TYPESTR,
                'strides': strides,
                'data': data,
                'version': 3}


class ByteSexpVector(NumpyArrayMixin, SexpVector):
    """Array of bytes.

    This is the R equivalent to a Python :class:`bytesarray`.
    """

    _R_TYPE = openrlib.rlib.RAWSXP
    _R_SIZEOF_ELT = _rinterface.ffi.sizeof('char')
    _NP_TYPESTR = '|u1'

    _R_GET_PTR = staticmethod(openrlib.RAW)

    @staticmethod
    def _CAST_IN(x: typing.Any) -> int:
        if isinstance(x, int):
            if x > 255:
                raise ValueError('byte must be in range(0, 256)')
            res = x
        elif isinstance(x, (bytes, bytearray)):
            if len(x) != 1:
                raise ValueError('byte must be a single character')
            res = ord(x)
        else:
            raise ValueError('byte must be an integer [0, 255] or a '
                             'single byte character')
        return res

    @staticmethod
    def _R_VECTOR_ELT(x, i: int) -> None:
        return openrlib.RAW(x)[i]

    @staticmethod
    def _R_SET_VECTOR_ELT(x, i: int, val) -> None:
        openrlib.RAW(x)[i] = val

    def __getitem__(self,
                    i: Union[int, slice]) -> Union[int, 'ByteSexpVector']:

        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            res = openrlib.RAW_ELT(cdata, i_c)
        elif isinstance(i, slice):
            res = type(self).from_iterable(
                [openrlib.RAW_ELT(
                    cdata, i_c
                ) for i_c in range(*i.indices(len(self)))
                ]
            )
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))
        return res

    def __setitem__(self, i: Union[int, slice], value) -> None:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            openrlib.RAW(cdata)[i_c] = self._CAST_IN(value)
        elif isinstance(i, slice):
            for i_c, v in zip(range(*i.indices(len(self))), value):
                if v > 255:
                    raise ValueError('byte must be in range(0, 256)')
                openrlib.RAW(cdata)[i_c] = self._CAST_IN(v)
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))


class BoolSexpVector(NumpyArrayMixin, SexpVector):
    """Array of booleans.

    Note that R is internally storing booleans as integers to
    allow an additional "NA" value to represent missingness."""

    _R_TYPE = openrlib.rlib.LGLSXP
    _R_SIZEOF_ELT = _rinterface.ffi.sizeof('Rboolean')
    _NP_TYPESTR = '|i'
    _R_VECTOR_ELT = openrlib.LOGICAL_ELT
    _R_SET_VECTOR_ELT = openrlib.SET_LOGICAL_ELT
    _R_GET_PTR = staticmethod(openrlib.LOGICAL)

    @staticmethod
    def _CAST_IN(x):
        if x is None or x == openrlib.rlib.R_NaInt:
            return NA_Logical
        else:
            return bool(x)

    def __getitem__(self, i: Union[int, slice]) -> Union[typing.Optional[bool],
                                                         'BoolSexpVector']:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            elt = openrlib.LOGICAL_ELT(cdata, i_c)
            res = na_values.NA_Logical if elt == NA_Logical else bool(elt)
        elif isinstance(i, slice):
            res = type(self).from_iterable(
                [openrlib.LOGICAL_ELT(cdata, i_c)
                 for i_c in range(*i.indices(len(self)))]
            )
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))
        return res

    def __setitem__(self, i: Union[int, slice], value) -> None:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            openrlib.SET_LOGICAL_ELT(cdata, i_c,
                                     int(value))
        elif isinstance(i, slice):
            for i_c, v in zip(range(*i.indices(len(self))), value):
                openrlib.SET_LOGICAL_ELT(cdata, i_c,
                                         int(v))
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))

    def memoryview(self) -> memoryview:
        return vector_memoryview(self, 'int', 'i')


def nullable_int(v):
    if type(v) is float and math.isnan(v):
        return openrlib.rlib.R_NaInt
    else:
        return int(v)


class IntSexpVector(NumpyArrayMixin, SexpVector):

    _R_TYPE = openrlib.rlib.INTSXP
    _R_SET_VECTOR_ELT = openrlib.SET_INTEGER_ELT
    _R_VECTOR_ELT = openrlib.INTEGER_ELT
    _R_SIZEOF_ELT = _rinterface.ffi.sizeof('int')
    _NP_TYPESTR = '|i'

    _R_GET_PTR = staticmethod(openrlib.INTEGER)
    _CAST_IN = staticmethod(nullable_int)

    def __getitem__(self, i: Union[int, slice]) -> Union[int, 'IntSexpVector']:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            res = openrlib.INTEGER_ELT(cdata, i_c)
            if res == NA_Integer:
                res = NA_Integer
        elif isinstance(i, slice):
            res = type(self).from_iterable(
                [openrlib.INTEGER_ELT(
                    cdata, i_c
                ) for i_c in range(*i.indices(len(self)))]
            )
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))
        return res

    def __setitem__(self, i: Union[int, slice], value) -> None:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            openrlib.SET_INTEGER_ELT(cdata, i_c,
                                     int(value))
        elif isinstance(i, slice):
            for i_c, v in zip(range(*i.indices(len(self))), value):
                openrlib.SET_INTEGER_ELT(cdata, i_c,
                                         int(v))
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))

    def memoryview(self) -> memoryview:
        return vector_memoryview(self, 'int', 'i')


class FloatSexpVector(NumpyArrayMixin, SexpVector):

    _R_TYPE = openrlib.rlib.REALSXP
    _R_VECTOR_ELT = openrlib.REAL_ELT
    _R_SET_VECTOR_ELT = openrlib.SET_REAL_ELT
    _R_SIZEOF_ELT = _rinterface.ffi.sizeof('double')
    _NP_TYPESTR = '|d'

    _CAST_IN = staticmethod(float)
    _R_GET_PTR = staticmethod(openrlib.REAL)

    def __getitem__(
            self, i: Union[int, slice]
    ) -> Union[float, 'FloatSexpVector']:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            res = openrlib.REAL_ELT(cdata, i_c)
        elif isinstance(i, slice):
            res = type(self).from_iterable(
                [openrlib.REAL_ELT(
                    cdata, i_c) for i_c in range(*i.indices(len(self)))]
            )
        else:
            raise TypeError('Indices must be integers or slices, not %s' %
                            type(i))
        return res

    def __setitem__(self, i: Union[int, slice], value) -> None:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            openrlib.SET_REAL_ELT(cdata, i_c,
                                  float(value))
        elif isinstance(i, slice):
            for i_c, v in zip(range(*i.indices(len(self))), value):
                openrlib.SET_REAL_ELT(cdata, i_c,
                                      float(v))
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))

    def memoryview(self) -> memoryview:
        return vector_memoryview(self, 'double', 'd')


class ComplexSexpVector(SexpVector):

    _R_TYPE = openrlib.rlib.CPLXSXP
    _R_GET_PTR = staticmethod(openrlib.COMPLEX)
    _R_SIZEOF_ELT = _rinterface.ffi.sizeof('Rcomplex')

    @staticmethod
    def _R_VECTOR_ELT(x, i):
        return openrlib.COMPLEX(x)[i]

    @staticmethod
    def _R_SET_VECTOR_ELT(x, i, v):
        openrlib.COMPLEX(x).__setitem__(i, v)

    @staticmethod
    def _CAST_IN(x):
        if isinstance(x, complex):
            res = (x.real, x.imag)
        else:
            try:
                res = (x.r, x.i)
            except AttributeError:
                raise TypeError(
                    'Unable to turn value into an R complex number.'
                )
        return res

    def __getitem__(
            self, i: Union[int, slice]
    ) -> Union[complex, 'ComplexSexpVector']:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            _ = openrlib.COMPLEX_ELT(cdata, i_c)
            res = complex(_.r, _.i)
        elif isinstance(i, slice):
            res = type(self).from_iterable(
                [openrlib.COMPLEX_ELT(
                    cdata, i_c) for i_c in range(*i.indices(len(self)))]
            )
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))
        return res

    def __setitem__(self, i: Union[int, slice], value) -> None:
        cdata = self.__sexp__._cdata
        if isinstance(i, int):
            i_c = _rinterface._python_index_to_c(cdata, i)
            openrlib.COMPLEX(cdata)[i_c] = self._CAST_IN(value)
        elif isinstance(i, slice):
            for i_c, v in zip(range(*i.indices(len(self))), value):
                openrlib.COMPLEX(cdata)[i_c] = self._CAST_IN(v)
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))


class ListSexpVector(SexpVector):
    """R list.

    An R list an R vector (array) that is similar to a Python list in
    the sense that items in the list can be of any type, whereas most
    other R vectors are homogeneous (all items are of the same type).
    """
    _R_TYPE = openrlib.rlib.VECSXP
    _R_GET_PTR = staticmethod(openrlib._VECTOR_PTR)
    _R_SIZEOF_ELT = None
    _R_VECTOR_ELT = openrlib.rlib.VECTOR_ELT
    _R_SET_VECTOR_ELT = openrlib.rlib.SET_VECTOR_ELT
    _CAST_IN = staticmethod(conversion._get_cdata)


class PairlistSexpVector(SexpVector):
    """R pairlist.

    A R pairlist is rarely used outside of R's internal libraries and
    a relatively small number of use cases. It is essentially a LISP-like
    list of (name, value) pairs.
    """
    _R_TYPE = openrlib.rlib.LISTSXP
    _R_GET_PTR = None
    _R_SIZEOF_ELT = None
    _R_VECTOR_ELT = None
    _R_SET_VECTOR_ELT = None
    _CAST_IN = staticmethod(conversion._get_cdata)

    def __getitem__(self, i: Union[int, slice]) -> Sexp:
        cdata = self.__sexp__._cdata
        rlib = openrlib.rlib
        if isinstance(i, int):
            # R-exts says that it is converted to a VECSXP when subsetted.
            i_c = _rinterface._python_index_to_c(cdata, i)
            item_cdata = rlib.Rf_nthcdr(cdata, i_c)
            with memorymanagement.rmemory() as rmemory:
                res_cdata = rmemory.protect(
                    rlib.Rf_allocVector(RTYPES.VECSXP, 1))
                rlib.SET_VECTOR_ELT(
                    res_cdata,
                    0,
                    rlib.CAR(
                        item_cdata
                    ))
                res_name = rmemory.protect(
                    rlib.Rf_allocVector(RTYPES.STRSXP, 1))
                item_cdata_name = rlib.PRINTNAME(rlib.TAG(item_cdata))
                if _rinterface._TYPEOF(item_cdata_name) != rlib.NILSXP:
                    rlib.SET_STRING_ELT(
                        res_name,
                        0,
                        item_cdata_name)
                    rlib.Rf_namesgets(res_cdata, res_name)
                res = conversion._cdata_to_rinterface(res_cdata)
        elif isinstance(i, slice):
            iter_indices = range(*i.indices(len(self)))
            n = len(iter_indices)
            with memorymanagement.rmemory() as rmemory:
                res_cdata = rmemory.protect(
                    rlib.Rf_allocVector(
                        self._R_TYPE, n)
                )
                iter_res_cdata = res_cdata
                prev_i = 0
                lst_cdata = self.__sexp__._cdata
                for i in iter_indices:
                    if i >= len(self):
                        raise IndexError('index out of range')
                    lst_cdata = rlib.Rf_nthcdr(lst_cdata, i - prev_i)
                    prev_i = i
                    rlib.SETCAR(iter_res_cdata,
                                rlib.CAR(lst_cdata))
                    rlib.SET_TAG(iter_res_cdata,
                                 rlib.TAG(lst_cdata))
                    iter_res_cdata = rlib.CDR(iter_res_cdata)
                res = conversion._cdata_to_rinterface(res_cdata)
        else:
            raise TypeError(
                'Indices must be integers or slices, not %s' % type(i))
        return res

    @classmethod
    @_cdata_res_to_rinterface
    def from_iterable(cls, iterable, cast_in=None):
        raise NotImplementedError()


class ExprSexpVector(SexpVector):
    _R_TYPE = openrlib.rlib.EXPRSXP
    _R_GET_PTR = None
    _CAST_IN = None
    _R_SIZEOF_ELT = None
    _R_VECTOR_ELT = openrlib.rlib.VECTOR_ELT
    _R_SET_VECTOR_ELT = None


class LangSexpVector(SexpVector):
    _R_TYPE = openrlib.rlib.LANGSXP
    _R_GET_PTR = None
    _CAST_IN = None
    _R_SIZEOF_ELT = None
    _R_VECTOR_ELT = None
    _R_SET_VECTOR_ELT = None

    @_cdata_res_to_rinterface
    def __getitem__(self, i: int):
        cdata = self.__sexp__._cdata
        i_c = _rinterface._python_index_to_c(cdata, i)
        return openrlib.rlib.CAR(
            openrlib.rlib.Rf_nthcdr(cdata, i_c)
        )

    def __setitem__(self, i: int, value: sexp.SupportsSEXP) -> None:
        cdata = self.__sexp__._cdata
        i_c = _rinterface._python_index_to_c(cdata, i)
        openrlib.rlib.SETCAR(
            openrlib.rlib.Rf_nthcdr(cdata, i_c),
            value.__sexp__._cdata
        )


class SexpClosure(Sexp):

    @_cdata_res_to_rinterface
    def __call__(self, *args, **kwargs) -> Sexp:
        error_occured = _rinterface.ffi.new('int *', 0)
        with memorymanagement.rmemory() as rmemory:
            call_r = rmemory.protect(
                _rinterface.build_rcall(self.__sexp__._cdata, args,
                                        kwargs.items()))
            call_context = _evaluation_context
            res = rmemory.protect(
                openrlib.rlib.R_tryEval(
                    call_r,
                    call_context.__sexp__._cdata,
                    error_occured)
            )
            if error_occured[0]:
                raise embedded.RRuntimeError(_rinterface._geterrmessage())
        return res

    @_cdata_res_to_rinterface
    def rcall(self, keyvals,
              environment: typing.Optional[SexpEnvironment] = None):
        """Call/evaluate an R function.

        Args:
        - keyvals: a sequence of key/value (name/parameter) pairs. A
          name/parameter that is None will indicated an unnamed parameter.
          Like in R, keys/names do not have to be unique, partial matching
          can be used, and named/unnamed parameters can occur at any position
          in the sequence.
        - environment: an optional R environment to evaluate the function.
        """
        # TODO: check keyvals are pairs ?
        if environment is None:
            environment = _evaluation_context
        assert isinstance(environment, SexpEnvironment)
        error_occured = _rinterface.ffi.new('int *', 0)

        with memorymanagement.rmemory() as rmemory:
            call_r = rmemory.protect(
                _rinterface.build_rcall(self.__sexp__._cdata, [],
                                        keyvals))
            res = rmemory.protect(
                openrlib.rlib.R_tryEval(call_r,
                                        environment.__sexp__._cdata,
                                        error_occured))
            if error_occured[0]:
                raise embedded.RRuntimeError(_rinterface._geterrmessage())
        return res

    @property
    @_cdata_res_to_rinterface
    def closureenv(self) -> SexpEnvironment:
        """Closure of the R function."""
        return openrlib.rlib.CLOENV(self.__sexp__._cdata)


class SexpS4(Sexp):
    """R "S4" object."""
    pass


# TODO: clean up
def make_extptr(obj, tag, protected):
    if protected is None:
        cdata_protected = openrlib.rlib.R_NilValue
    else:
        try:
            cdata_protected = protected.__sexp__._cdata
        except AttributeError:
            raise TypeError('Argument protected must inherit from %s' %
                            type(Sexp))

    ptr = _rinterface.ffi.new_handle(obj)
    with memorymanagement.rmemory() as rmemory:
        cdata = rmemory.protect(
            openrlib.rlib.R_MakeExternalPtr(
                ptr,
                tag,
                cdata_protected))
        openrlib.rlib.R_RegisterCFinalizer(
            cdata,
            (_rinterface._capsule_finalizer_c
             if _rinterface._capsule_finalizer_c
             else _rinterface._capsule_finalizer))
        res = _rinterface.SexpCapsuleWithPassenger(cdata, obj, ptr)
    return res


class SexpExtPtr(Sexp):

    TYPE_TAG = 'Python'

    @classmethod
    def from_pyobject(cls, func, tag: str = TYPE_TAG,
                      protected=None):
        if not isinstance(tag, str):
            raise TypeError('The tag must be a string.')
        scaps = make_extptr(func,
                            conversion._str_to_charsxp(cls.TYPE_TAG),
                            protected)
        res = cls(scaps)
        if tag != cls.TYPE_TAG:
            res.TYPE_TAG = tag
        return res


# TODO: Only use rinterface-level ?
conversion._R_RPY2_MAP.update({
    openrlib.rlib.NILSXP: NULLType,
    openrlib.rlib.EXPRSXP: ExprSexpVector,
    openrlib.rlib.LANGSXP: LangSexpVector,
    openrlib.rlib.ENVSXP: SexpEnvironment,
    openrlib.rlib.RAWSXP: ByteSexpVector,
    openrlib.rlib.LGLSXP: BoolSexpVector,
    openrlib.rlib.INTSXP: IntSexpVector,
    openrlib.rlib.REALSXP: FloatSexpVector,
    openrlib.rlib.CPLXSXP: ComplexSexpVector,
    openrlib.rlib.STRSXP: StrSexpVector,
    openrlib.rlib.VECSXP: ListSexpVector,
    openrlib.rlib.LISTSXP: PairlistSexpVector,
    openrlib.rlib.CLOSXP: SexpClosure,
    openrlib.rlib.BUILTINSXP: SexpClosure,
    openrlib.rlib.SPECIALSXP: SexpClosure,
    openrlib.rlib.EXTPTRSXP: SexpExtPtr,
    openrlib.rlib.SYMSXP: SexpSymbol,
    openrlib.rlib.S4SXP: SexpS4
    })
conversion._R_RPY2_DEFAULT_MAP = Sexp

conversion._PY_RPY2_MAP.update({
    int: conversion._int_to_sexp,
    float: conversion._float_to_sexp,
    complex: conversion._complex_to_sexp
    })

conversion._PY_R_MAP.update({
    _rinterface.ffi.CData: False,
    # integer
    int: conversion._int_to_sexp,
    sexp.NAIntegerType: conversion._int_to_sexp,
    # float
    float: conversion._float_to_sexp,
    sexp.NARealType: conversion._float_to_sexp,
    # boolean
    bool: conversion._bool_to_sexp,
    sexp.NALogicalType: conversion._bool_to_sexp,
    # string
    str: conversion._str_to_sexp,
    sexp.CharSexp: None,
    sexp.NACharacterType: None,
    # complex
    complex: conversion._complex_to_sexp,
    sexp.NAComplexType: conversion._complex_to_sexp,
    # None
    type(None): lambda x: openrlib.rlib.R_NilValue})


def vector(iterable, rtype: RTYPES) -> SexpVector:
    """Create an R vector.

    While the different types of R vectors all have their own class,
    the creation of array in Python is often available through factory
    function that accept the type of the array as a parameters. This
    function is providing a similar functionality for R vectors."""
    error = False
    try:
        cls = conversion._R_RPY2_MAP[rtype]
    except KeyError:
        error = True
    if not error and not issubclass(cls, SexpVector):
        error = True
    if error:
        raise ValueError(
            'Unable to build a vector from type "%s"' % RTYPES(rtype))
    return cls.from_iterable(iterable)


class RRuntimeWarning(RuntimeWarning):
    pass


NULL = None
MissingArg = _MissingArgType()
NA_Character = None
NA_Integer = None
NA_Logical = None
NA = None
NA_Real = None
NA_Complex = None


def initr_simple() -> typing.Optional[int]:
    """Initialize R's embedded C library."""
    with openrlib.rlock:
        status = embedded._initr()
        atexit.register(endr, 0)
        _rinterface._register_external_symbols()
        _post_initr_setup()
        return status


def initr_checkenv() -> typing.Optional[int]:
    # Force the internal initialization flag if there is an environment
    # variable that indicates that R was already initialized in the current
    # process.

    status = None
    with openrlib.rlock:
        if embedded.is_r_externally_initialized():
            embedded._setinitialized()
        else:
            status = embedded._initr()
            embedded.set_python_process_info()
        _rinterface._register_external_symbols()
        _post_initr_setup()
    return status


initr = initr_checkenv


def _update_R_ENC_PY():
    conversion._R_ENC_PY[openrlib.rlib.CE_UTF8] = 'utf-8'

    l10n_info = tuple(baseenv['l10n_info']())
    if platform.system() == 'Windows':
        val_latin1 = 'cp{:d}'.format(l10n_info[3][0])
    else:
        val_latin1 = 'latin1'
    conversion._R_ENC_PY[openrlib.rlib.CE_LATIN1] = val_latin1

    if l10n_info[1]:
        val_native = conversion._R_ENC_PY[openrlib.rlib.CE_UTF8]
    else:
        val_native = val_latin1
    conversion._R_ENC_PY[openrlib.rlib.CE_NATIVE] = val_native

    conversion._R_ENC_PY[openrlib.rlib.CE_ANY] = 'utf-8'


def _post_initr_setup() -> None:

    emptyenv.__sexp__ = _rinterface.SexpCapsule(openrlib.rlib.R_EmptyEnv)
    baseenv.__sexp__ = _rinterface.SexpCapsule(openrlib.rlib.R_BaseEnv)
    globalenv.__sexp__ = _rinterface.SexpCapsule(openrlib.rlib.R_GlobalEnv)

    global NULL
    NULL = NULLType()

    MissingArg._sexpobject = _rinterface.UnmanagedSexpCapsule(
        openrlib.rlib.R_MissingArg
    )

    global NA_Character
    na_values.NA_Character = sexp.NACharacterType()
    NA_Character = na_values.NA_Character

    global NA_Integer
    na_values.NA_Integer = sexp.NAIntegerType(openrlib.rlib.R_NaInt)
    NA_Integer = na_values.NA_Integer

    global NA_Logical, NA
    na_values.NA_Logical = sexp.NALogicalType(openrlib.rlib.R_NaInt)
    NA_Logical = na_values.NA_Logical
    NA = NA_Logical

    global NA_Real
    na_values.NA_Real = sexp.NARealType(openrlib.rlib.R_NaReal)
    NA_Real = na_values.NA_Real

    global NA_Complex
    na_values.NA_Complex = sexp.NAComplexType(
        _rinterface.ffi.new(
            'Rcomplex *',
            [openrlib.rlib.R_NaReal, openrlib.rlib.R_NaReal])
    )
    NA_Complex = na_values.NA_Complex

    warn_about_thread = False
    if threading.current_thread() is threading.main_thread():
        try:
            signal.signal(signal.SIGINT, _sigint_handler)
        except ValueError as ve:
            if str(ve) == 'signal only works in main thread':
                warn_about_thread = True
            else:
                raise ve
    else:
        warn_about_thread = True
    if warn_about_thread:
        warnings.warn(
            textwrap.dedent(
                """R is not initialized by the main thread.
                Its taking over SIGINT cannot be reversed here, and as a
                consequence the embedded R cannot be interrupted with Ctrl-C.
                Consider (re)setting the signal handler of your choice from
                the main thread."""
            )
        )

    _update_R_ENC_PY()


def rternalize(function: typing.Callable) -> SexpClosure:
    """ Make a Python function callable from R.

    Takes an arbitrary Python function and wrap it in such a way that
    it can be called from the R side.

    :param function: A Python callable object.
    :return: A wrapped R object that can be use like any other rpy2
    object.
    """

    assert callable(function)
    rpy_fun = SexpExtPtr.from_pyobject(function)
    # TODO: this is a hack. Find a better way.
    template = parse("""
      function(...) { .External(".Python", foo, ...);
    }
    """)
    template[0][2][1][2] = rpy_fun
    # TODO: use lower-level eval ?
    res = baseenv['eval'](template)
    # TODO: hack to prevent the nested function from having its
    #   refcount down to zero when returning
    res.__nested_sexp__ = rpy_fun.__sexp__
    return res


def process_revents() -> None:
    """Process R events.

    Calling this function a regular interval can help ensure that
    R is processing input events (e.g., resizing of a window with
    graphics)."""
    openrlib.rlib.rpy2_runHandlers(openrlib.rlib.R_InputHandlers)



import pytest
import rpy2.rlike.indexing as rfi


def test_order():
    seq  = (2, 1, 5, 3, 4)
    expected = (1, 2, 3, 4, 5)
    res = rfi.order(seq)
    for va, vb in zip(expected, res):
        assert seq[vb] == va



import jinja2

from rpy2.robjects import (vectors,
                           RObject,
                           SignatureTranslatedFunction,
                           RS4)
from rpy2 import rinterface
from rpy2.robjects.packages import SourceCode
from rpy2.robjects.packages import wherefrom
from IPython import get_ipython


template_list = jinja2.Template("""
<p><emph>{{ clsname }}</emph> with {{ rlist | length }} elements:</p>
<dl class="rpy2">
{%- for elt_i in range(display_neltmax) %}
      <dt>{{ rlist.names[elt_i] }}</dt>
      <dd>{{ rlist[elt_i] }}</dd>
{%- endfor %}
{%- if display_neltmax < (rlist | length) %}
      <dt>...</dt>
      <dd></dd>
{%- endif %}
</dl>
""")

template_vector_horizontal = jinja2.Template("""
<emph>{{ clsname }}</emph> with {{ vector | length }} elements:
<table class="{{ table_class }}">
<thead>
</thead>
<tbody>
<tr>
  {%- for elt_i in range(display_ncolmax - size_tail) %}
  <td>{{ vector[elt_i] }}</td>
  {%- endfor %}
  {%- if display_ncolmax < (vector | length) %}
  <td>...</td>
  {%- endif %}
  {%- for elt_i in elt_i_tail %}
      <td>{{ vector[elt_i] }}</td>
  {%- endfor %}
</tr>
</tbody>
</table>
""")

template_vector_vertical = jinja2.Template("""
<emph>{{ clsname }}</emph> with {{ vector | length }} elements:
<table class="{{ table_class }}">
<thead>
</thead>
<tbody>
  {%- for elt_i in range(display_nrowmax - size_tail) %}
  <tr>
    <td class="rpy2_rowname">{{ elt_i }}</td>
  {%- if has_vector_names %}
    <td class="rpy2_names">{{ vector.names[elt_i] }}</td>
  {%- endif %}
    <td>{{ "%s" | format(vector[elt_i]) | truncate(12) }}</td>
  </tr>
  {%- endfor %}
  {%- if display_nrowmax < (vector | length) %}
  <tr>
    <td class="rpy2_rowname">...</td>
  {%- if has_vector_names %}
    <td class="rpy2_names">...</td>
  {%- endif %}
    <td>...</td>
  </tr>
  {%- endif %}
  {%- for elt_i in elt_i_tail %}
  <tr>
    <td class="rpy2_rowname">{{ elt_i }}</td>
  {%- if has_vector_names %}
    <td class="rpy2_names">{{ vector.names[elt_i] }}</td>
  {%- endif %}
    <td>{{ "%s" | format(vector[elt_i]) | truncate(12) }}</td>
  </tr>
  {%- endfor %}
</tr>
</tbody>
</table>
""")

template_dataframe = jinja2.Template("""
<emph>{{ clsname }}</emph> with {{ dataf.nrow }} rows and
  {{ dataf | length }} columns:
<table class="{{ table_class }}">
  <thead>
    <tr class="rpy2_names">
      <th></th>
      {%- if has_rownames %}
      <th></th>
      {%- endif %}

{%- for col_i in range(display_ncolmax - size_coltail) %}
      <th>{{ dataf.names[col_i] }}</th>
{%- endfor %}
{%- if display_ncolmax < dataf.ncol %}
      <th>...</th>
{%- endif %}
{%- for col_i in col_i_tail %}
      <th>{{ dataf.names[col_i] }}</th>
{%- endfor %}
    </tr>
  </thead>
  <tbody>
{%- for row_i in range(display_nrowmax - size_rowtail) %}
    <tr>
      <td class="rpy2_rowname">{{ row_i }}</td>
      {%- if has_rownames %}
        <td class="rpy2_names">{{ dataf.rownames[row_i] }}</td>
      {%- endif %}
  {%- for col_i in range(display_ncolmax - size_coltail) %}
      <td>{{ dataf[col_i][row_i] }}</td>
  {%- endfor %}
  {%- if display_ncolmax < dataf.ncol %}
       <td>...</td>
  {%- endif %}
  {%- for col_i in col_i_tail %}
      <td>{{ dataf[col_i][row_i] }}</td>
  {%- endfor %}
    </tr>
{%- endfor %}

{%- if dataf.nrow > display_nrowmax %}
    <tr>
      <td class="rpy2_rowname">...</td>
      {%- if has_rownames %}
        <td class="rpy2_names">...</td>
      {%- endif %}

  {%- for col_i in range(display_ncolmax - size_coltail) %}
      <td>...</td>
  {%- endfor %}
  {%- if display_ncolmax < dataf.ncol %}
       <td>...</td>
  {%- endif %}
  {%- for col_i in range(2) %}
      <td>...</td>
  {%- endfor %}
    </tr>
{%- endif %}

{%- for row_i in row_i_tail %}
    <tr>
      <td class="rpy2_rowname">{{ row_i }}</td>
      {%- if has_rownames %}
        <td class="rpy2_names">{{ dataf.rownames[row_i] }}</td>
      {%- endif %}
  {%- for col_i in range(display_ncolmax - size_coltail) %}
      <td>{{ dataf[col_i][row_i] }}</td>
  {%- endfor %}
  {%- if display_ncolmax < dataf.ncol %}
       <td>...</td>
  {%- endif %}
  {%- for col_i in col_i_tail %}
      <td>{{ dataf[col_i][row_i] }}</td>
  {%- endfor %}
    </tr>
{%- endfor %}
  </tbody>
</table>
""")

template_ridentifiedobject = jinja2.Template("""
<ul style="list-style-type: none;">
<li>{{ clsname }} object</li>
<li>Origin in R: {{ origin }}</li>
<li>Class(es) in R:
  <ul>
  {%- for rclsname in obj.rclass %}
    <li>{{ rclsname }}</li>
  {%- endfor %}
  </ul>
</li>
</ul>
""")

template_rs4 = jinja2.Template("""
<table class="{{ table_class }}">
<tr>
  <th colspan="2">{{ clsname }} object</th>
</tr>
<tr>
  <th>Origin in R:</th>
  <td> {{ origin }}</td>
</tr>
<tr>
  <th>Class(es) in R:</th>
  <td>
    <ul>
  {%- for rclsname in obj.rclass %}
      <li>{{ rclsname }}</li>
  {%- endfor %}
    </ul>
  </td>
</tr>
<tr>
  <th>Attributes:</th>
  <td>
  <ul>
  {%- for sln in obj.slotnames() %}
    <li>{{ sln }}</li>
  {%- endfor %}
  </ul>
  </td>
</tr>
</table>
""")

template_sourcecode = jinja2.Template("""
<p>
R source code:
</p>
<style>
{{ syntax_highlighting }}
</style>
{{ sourcecode }}
""")


class StrFactorVector(vectors.FactorVector):

    def __getitem__(self, item):
        integer = super(StrFactorVector, self).__getitem__(item)
        # R is one-offset, Python is zero-offset
        return self.levels[integer-1]


class StrDataFrame(vectors.DataFrame):

    def __getitem__(self, item):
        obj = super(StrDataFrame, self).__getitem__(item)
        if isinstance(obj, vectors.FactorVector):
            obj = StrFactorVector(obj)
        return obj


def html_vector_horizontal(vector,
                           display_ncolmax=10,
                           size_tail=2,
                           table_class='rpy2_table'):
    if isinstance(vector, vectors.FactorVector):
        vector = StrFactorVector(vector)
    html = template_vector_horizontal.render({
        'table_class': table_class,
        'clsname': type(vector).__name__,
        'vector': vector,
        'display_ncolmax': min(display_ncolmax, len(vector)),
        'size_tail': size_tail,
        'elt_i_tail': range(max(0, len(vector)-size_tail), len(vector))})
    return html


def html_rlist(vector,
               display_nrowmax=10,
               size_tail=2,
               table_class='rpy2_table'):
    rg = range(max(0, len(vector)-size_tail), len(vector))
    html = template_vector_vertical.render({
        'table_class': table_class,
        'clsname': type(vector).__name__,
        'vector': vector,
        'has_vector_names': vector.names is not rinterface.NULL,
        'display_nrowmax': min(display_nrowmax, len(vector)),
        'size_tail': size_tail,
        'elt_i_tail': rg})
    return html


def html_rdataframe(dataf,
                    display_nrowmax=10,
                    display_ncolmax=6,
                    size_coltail=2,
                    size_rowtail=2,
                    table_class='rpy2_table'):
    html = template_dataframe.render(
        {'dataf': StrDataFrame(dataf),
         'table_class': table_class,
         'has_rownames': dataf.rownames is not None,
         'clsname': type(dataf).__name__,
         'display_nrowmax': min(display_nrowmax, dataf.nrow),
         'display_ncolmax': min(display_ncolmax, dataf.ncol),
         'col_i_tail': range(max(0, dataf.ncol-size_coltail), dataf.ncol),
         'row_i_tail': range(max(0, dataf.nrow-size_rowtail), dataf.nrow),
         'size_coltail': size_coltail,
         'size_rowtail': size_rowtail})
    return html


def html_sourcecode(sourcecode):
    from pygments import highlight
    from pygments.lexers import SLexer
    from pygments.formatters import HtmlFormatter
    formatter = HtmlFormatter()
    htmlcode = highlight(sourcecode, SLexer(), formatter)
    d = {'sourcecode': htmlcode,
         'syntax_highlighting': formatter.get_style_defs()}
    html = template_sourcecode.render(d)
    return html


def _dict_ridentifiedobject(obj):
    if hasattr(obj, '__rname__') and obj.__rname__ is not None:
        env = wherefrom(obj.__rname__)
        try:
            origin = env.do_slot('name')[0]
        except LookupError:
            origin = 'package:base ?'
    else:
        origin = '???'
    d = {'clsname': type(obj).__name__,
         'origin': origin,
         'obj': obj}
    return d


def html_ridentifiedobject(obj):
    d = _dict_ridentifiedobject(obj)
    html = template_ridentifiedobject.render(d)
    return html


def html_rs4(obj, table_class='rpy2_table'):
    d = _dict_ridentifiedobject(obj)
    d['table_class'] = table_class
    html = template_rs4.render(d)
    return html


def init_printing():
    ip = get_ipython()
    html_f = ip.display_formatter.formatters['text/html']
    html_f.for_type(vectors.Vector, html_vector_horizontal)
    html_f.for_type(vectors.ListVector, html_rlist)
    html_f.for_type(vectors.DataFrame, html_rdataframe)
    html_f.for_type(RObject, html_ridentifiedobject)
    html_f.for_type(RS4, html_rs4)
    html_f.for_type(SignatureTranslatedFunction, html_ridentifiedobject)
    html_f.for_type(SourceCode, html_sourcecode)



import pytest
import rpy2.rinterface
import rpy2.rinterface_lib.embedded
from threading import Thread
from rpy2.rinterface import embedded


class ThreadWithExceptions(Thread):
    """Wrapper around Thread allowing to record exceptions from the thread."""

    def run(self):
        self.exception = None
        try:
            self._target()
        except Exception as e:
            self.exception = e


@pytest.mark.skipif(embedded.rpy2_embeddedR_isinitialized,
                    reason='Can only be tested before R is initialized.')
def test_threading__initr():
    thread = ThreadWithExceptions(target=rpy2.rinterface_lib.embedded._initr)
    thread.start()
    thread.join()
    assert not thread.exception


@pytest.mark.skipif(embedded.rpy2_embeddedR_isinitialized,
                    reason='Can only be tested before R is initialized.')
def test_threading_initr_simple():
    # This initialization performs more post-initialization setup compared to _initr.
    # It will check whether R is initialized from the main thread.
    thread = ThreadWithExceptions(target=rpy2.rinterface.initr_simple)
    with pytest.warns(
        UserWarning,
        match=(
            r'R is not initialized by the main thread.\n'
            r'\W+Its taking over SIGINT cannot be reversed here.+'
        )
    ):
        thread.start()
        thread.join()
    assert not thread.exception



"""This module handles the conversion of data structures
between R objects handled by rpy2 and pandas objects."""

import rpy2.robjects.conversion as conversion
import rpy2.rinterface as rinterface
from rpy2.rinterface_lib import na_values
from rpy2.rinterface import (IntSexpVector,
                             ListSexpVector,
                             Sexp,
                             SexpVector,
                             StrSexpVector)
import datetime
import functools
import math
import numpy
import pandas
import pandas.core.series
from pandas.core.frame import DataFrame as PandasDataFrame
from pandas.core.dtypes.api import is_datetime64_any_dtype
import warnings

from collections import OrderedDict
from rpy2.robjects.vectors import (BoolVector,
                                   DataFrame,
                                   DateVector,
                                   FactorVector,
                                   FloatSexpVector,
                                   StrVector,
                                   IntVector,
                                   POSIXct)

# The pandas converter requires numpy.
import rpy2.robjects.numpy2ri as numpy2ri

original_converter = None

ISOdatetime = rinterface.baseenv['ISOdatetime']
as_vector = rinterface.baseenv['as.vector']

converter = conversion.Converter('original pandas conversion')
py2rpy = converter.py2rpy
rpy2py = converter.rpy2py

# numpy types for Pandas columns that require (even more) special handling
dt_O_type = numpy.dtype('O')

# pandas types for series of integer (optional missing) values.
integer_array_types = ('Int8', 'Int16', 'Int32', 'Int64', 'UInt8',
                       'UInt16', 'UInt32', 'UInt64')


@py2rpy.register(PandasDataFrame)
def py2rpy_pandasdataframe(obj):
    od = OrderedDict()
    for name, values in obj.iteritems():
        try:
            od[name] = conversion.py2rpy(values)
        except Exception as e:
            warnings.warn('Error while trying to convert '
                          'the column "%s". Fall back to string conversion. '
                          'The error is: %s'
                          % (name, str(e)))
            od[name] = StrVector(values)

    return DataFrame(od)


@py2rpy.register(pandas.Index)
def py2rpy_pandasindex(obj):
    if obj.dtype.kind == 'O':
        return StrVector(obj)
    else:
        # pandas2ri should definitely not have to know which paths remain to be
        # converted by numpy2ri
        # Answer: the thing is that pandas2ri builds on the conversion
        # rules defined by numpy2ri - deferring to numpy2ri is allowing
        # us to reuse that code.
        return numpy2ri.numpy2rpy(obj)


def py2rpy_categoryseries(obj):
    for c in obj.cat.categories:
        if not isinstance(c, str):
            raise ValueError('Converting pandas "Category" series to '
                             'R factor is only possible when categories '
                             'are strings.')
    res = IntSexpVector(list(rinterface.NA_Integer if x == -1 else x+1
                             for x in obj.cat.codes))
    res.do_slot_assign('levels', StrSexpVector(obj.cat.categories))
    if obj.cat.ordered:
        res.rclass = StrSexpVector(('ordered', 'factor'))
    else:
        res.rclass = StrSexpVector(('factor',))
    return res


def _str_populate_r_vector(iterable, r_vector,
                           set_elt,
                           cast_value):
    for i, v in enumerate(iterable):
        if (
                v is None
                or
                v is pandas.NA
                or
                (isinstance(v, float) and math.isnan(v))
        ):
            v = na_values.NA_Character
        set_elt(r_vector, i, cast_value(v))


def _int_populate_r_vector(iterable, r_vector,
                           set_elt,
                           cast_value):
    for i, v in enumerate(iterable):
        if v is None or v is pandas.NA:
            v = math.nan
        set_elt(r_vector, i, cast_value(v))


_PANDASTYPE2RPY2 = {
    datetime.date: DateVector,
    int: functools.partial(
        IntVector.from_iterable,
        populate_func=_int_populate_r_vector
    ),
    bool: BoolVector,
    None: BoolVector,
    str: functools.partial(
        StrVector.from_iterable,
        populate_func=_str_populate_r_vector
    ),
    bytes: (numpy2ri.converter.py2rpy.registry[
        numpy.ndarray
    ])
}


@py2rpy.register(pandas.core.series.Series)
def py2rpy_pandasseries(obj):
    if obj.dtype.name == 'O':
        warnings.warn('Element "%s" is of dtype "O" and converted '
                      'to R vector of strings.' % obj.name)
        res = StrVector(obj)
    elif obj.dtype.name == 'category':
        res = py2rpy_categoryseries(obj)
        res = FactorVector(res)
    elif is_datetime64_any_dtype(obj.dtype):
        # time series
        tzname = obj.dt.tz.zone if obj.dt.tz else ''
        d = [IntVector([x.year for x in obj]),
             IntVector([x.month for x in obj]),
             IntVector([x.day for x in obj]),
             IntVector([x.hour for x in obj]),
             IntVector([x.minute for x in obj]),
             FloatSexpVector([x.second + x.microsecond * 1e-6 for x in obj])]
        res = ISOdatetime(*d, tz=StrSexpVector([tzname]))
        # TODO: can the POSIXct be created from the POSIXct constructor ?
        # (is '<M8[ns]' mapping to Python datetime.datetime ?)
        res = POSIXct(res)
    elif obj.dtype.type == str:
        res = _PANDASTYPE2RPY2[str](obj)
    elif obj.dtype.name in integer_array_types:
        res = _PANDASTYPE2RPY2[int](obj)
        if len(obj.shape) == 1:
            if obj.dtype != dt_O_type:
                # force into an R vector
                res = as_vector(res)
    elif (obj.dtype == dt_O_type):
        homogeneous_type = None
        for x in obj.values:
            if x is None:
                continue
            if homogeneous_type is None:
                homogeneous_type = type(x)
                continue
            if ((type(x) is not homogeneous_type)
                and not
                ((isinstance(x, float) and math.isnan(x))
                 or pandas.isna(x))):
                raise ValueError('Series can only be of one type, or None '
                                 '(and here we have %s and %s).' %
                                 (homogeneous_type, type(x)))
        # TODO: Could this be merged with obj.type.name == 'O' case above ?
        res = _PANDASTYPE2RPY2[homogeneous_type](obj)
    else:
        # converted as a numpy array
        func = numpy2ri.converter.py2rpy.registry[numpy.ndarray]
        # current conversion as performed by numpy

        res = func(obj.values)
        if len(obj.shape) == 1:
            if (obj.dtype != dt_O_type):
                # force into an R vector
                res = as_vector(res)

    # "index" is equivalent to "names" in R
    if obj.ndim == 1:
        res.do_slot_assign('names',
                           StrVector(tuple(str(x) for x in obj.index)))
    else:
        res.do_slot_assign('dimnames',
                           SexpVector(conversion.py2rpy(obj.index)))
    return res


@rpy2py.register(SexpVector)
def ri2py_vector(obj):
    res = numpy2ri.rpy2py(obj)
    return res


@rpy2py.register(IntSexpVector)
def rpy2py_intvector(obj):
    # special case for factors
    if 'factor' in obj.rclass:
        codes = [x-1 if x > 0 else -1 for x in numpy.array(obj)]
        res = pandas.Categorical.from_codes(
            codes,
            categories=list(obj.do_slot('levels')),
            ordered='ordered' in obj.rclass
        )
    else:
        res = numpy2ri.rpy2py(obj)
    return res


@rpy2py.register(FloatSexpVector)
def rpy2py_floatvector(obj):
    if POSIXct.isrinstance(obj):
        return rpy2py(POSIXct(obj))
    else:
        return numpy2ri.rpy2py(obj)


@rpy2py.register(POSIXct)
def rpy2py_posixct(obj):
    return pandas.to_datetime(tuple(obj.iter_localized_datetime()),
                              errors='coerce')


@rpy2py.register(ListSexpVector)
def rpy2py_listvector(obj):
    if 'data.frame' in obj.rclass:
        res = rpy2py(DataFrame(obj))
    else:
        res = numpy2ri.rpy2py(obj)
    return res


@rpy2py.register(DataFrame)
def rpy2py_dataframe(obj):
    items = OrderedDict((k, rpy2py(v) if isinstance(v, Sexp) else v)
                        for k, v in obj.items())
    res = PandasDataFrame.from_dict(items)
    res.index = obj.rownames
    return res


def activate():
    warnings.warn('The global conversion available with activate() '
                  'is deprecated and will be removed in the next '
                  'major release. Use a local converter.',
                  category=DeprecationWarning)
    global original_converter
    # If module is already activated, there is nothing to do.
    if original_converter is not None:
        return

    original_converter = conversion.Converter(
        'snapshot before pandas conversion',
        template=conversion.converter)
    numpy2ri.activate()
    new_converter = conversion.Converter('snapshot before pandas conversion',
                                         template=conversion.converter)
    numpy2ri.deactivate()

    for k, v in py2rpy.registry.items():
        if k is object:
            continue
        new_converter.py2rpy.register(k, v)

    for k, v in rpy2py.registry.items():
        if k is object:
            continue
        new_converter.rpy2py.register(k, v)

    conversion.set_conversion(new_converter)


def deactivate():
    global original_converter

    # If module has never been activated or already deactivated,
    # there is nothing to do
    if original_converter is None:
        return

    conversion.set_conversion(original_converter)
    original_converter = None



import copy
import gc
import pytest
import rpy2.rinterface as rinterface

rinterface.initr()


def test_invalid_init():
    with pytest.raises(ValueError):
        rinterface.Sexp('a')


def test_init_from_existing():
    sexp = rinterface.baseenv.find('letters')
    sexp_new = rinterface.Sexp(sexp)
    assert sexp_new._sexpobject is sexp._sexpobject


def test_typeof():
    assert isinstance(rinterface.baseenv.typeof, int)


def test_get():
    sexp = rinterface.baseenv.find('letters')
    assert sexp.typeof == rinterface.RTYPES.STRSXP

    sexp = rinterface.baseenv.find('pi')
    assert sexp.typeof == rinterface.RTYPES.REALSXP

    sexp = rinterface.baseenv.find('options')
    assert sexp.typeof == rinterface.RTYPES.CLOSXP


@pytest.mark.parametrize('cls',
                         (rinterface.IntSexpVector, rinterface.ListSexpVector))
def test_list_attrs(cls):
    x = cls((1, 2, 3))
    assert len(x.list_attrs()) == 0
    x.do_slot_assign('a', rinterface.IntSexpVector((33, )))
    assert len(x.list_attrs()) == 1
    assert 'a' in x.list_attrs()


def test_do_slot():
    sexp = rinterface.baseenv.find('.Platform')
    names = sexp.do_slot('names')
    assert len(names) > 1
    assert 'OS.type' in names


def test_names():
    sexp = rinterface.baseenv.find('.Platform')
    names = sexp.names
    assert len(names) > 1
    assert 'OS.type' in names


def test_names_set():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    assert sexp.names.rid == rinterface.NULL.rid
    sexp.names = rinterface.StrSexpVector(['a', 'b', 'c'])
    assert len(sexp.names) > 1
    assert tuple(sexp.names) == ('a', 'b', 'c')


def test_names_set_invalid():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    assert sexp.names.rid == rinterface.NULL.rid
    with pytest.raises(ValueError):
        sexp.names = ('a', 'b', 'c')


def test_do_slot_missing():
    sexp = rinterface.baseenv.find('pi')
    with pytest.raises(LookupError):
        sexp.do_slot('foo')


def test_do_slot_not_string():
    sexp = rinterface.baseenv.find('pi')
    with pytest.raises(ValueError):
        sexp.do_slot(None)


def test_do_slot_empty_string():
    sexp = rinterface.baseenv.find('pi')
    with pytest.raises(ValueError):
        sexp.do_slot('')


def test_do_slot_assign_create():
    sexp = rinterface.IntSexpVector([])
    slot_value = rinterface.IntSexpVector([3, ])
    sexp.do_slot_assign('foo', slot_value)
    slot_value_back = sexp.do_slot('foo')
    assert len(slot_value_back) == len(slot_value)
    assert all(x == y for x, y in zip(slot_value, slot_value_back))


def test_do_slot_reassign():
    sexp = rinterface.IntSexpVector([])
    slot_value_a = rinterface.IntSexpVector([3, ])
    sexp.do_slot_assign('foo', slot_value_a)
    slot_value_b = rinterface.IntSexpVector([5, 6])
    sexp.do_slot_assign('foo', slot_value_b)
    slot_value_back = sexp.do_slot('foo')
    assert len(slot_value_b) == len(slot_value_back)
    assert all(x == y for x, y in zip(slot_value_b, slot_value_back))


def test_do_slot_assign_empty_string():
    sexp = rinterface.IntSexpVector([])
    slot_value = rinterface.IntSexpVector([3, ])
    with pytest.raises(ValueError):
        sexp.do_slot_assign('', slot_value)


def test_sexp_rsame_true():
    sexp_a = rinterface.baseenv.find("letters")
    sexp_b = rinterface.baseenv.find("letters")
    assert sexp_a.rsame(sexp_b)


def test_sexp_rsame_false():
    sexp_a = rinterface.baseenv.find("letters")
    sexp_b = rinterface.baseenv.find("pi")
    assert not sexp_a.rsame(sexp_b)


def test_sexp_rsame_invalid():
    sexp_a = rinterface.baseenv.find("letters")
    with pytest.raises(ValueError):
        sexp_a.rsame('foo')


def test___sexp__():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    sexp_count = sexp.__sexp_refcount__
    sexp_cobj = sexp.__sexp__
    d = dict(rinterface._rinterface.protected_rids())
    assert sexp_count == d[sexp.rid]
    assert sexp_count == sexp.__sexp_refcount__
    sexp2 = rinterface.IntSexpVector([4, 5, 6, 7])
    sexp2_rid = sexp2.rid
    sexp2.__sexp__ = sexp_cobj
    del(sexp)
    gc.collect()
    d = dict(rinterface._rinterface.protected_rids())
    assert d.get(sexp2_rid) is None


def test_rclass_get():
    sexp = rinterface.baseenv.find('character')(1)
    assert len(sexp.rclass) == 1
    assert sexp.rclass[0] == 'character'

    sexp = rinterface.baseenv.find('matrix')(0)
    if rinterface.evalr('R.version$major')[0] >= '4':
        assert tuple(sexp.rclass) == ('matrix', 'array')
    else:
        assert tuple(sexp.rclass) == ('matrix', )

    sexp = rinterface.baseenv.find('array')(0)
    assert len(sexp.rclass) == 1
    assert sexp.rclass[0] == 'array'

    sexp = rinterface.baseenv.find('new.env')()
    assert len(sexp.rclass) == 1
    assert sexp.rclass[0] == 'environment'


def test_rclass_get_sym():
    # issue #749
    fit = rinterface.evalr("""
    stats::lm(y ~ x, data=base::data.frame(y=1:10, x=2:11))
    """)
    assert tuple(fit[9].rclass) == ('call', )


def test_rclass_set():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    sexp.rclass = rinterface.StrSexpVector(['foo'])
    assert len(sexp.rclass) == 1
    assert sexp.rclass[0] == 'foo'

    sexp.rclass = 'bar'
    assert len(sexp.rclass) == 1
    assert sexp.rclass[0] == 'bar'


def test_rclass_set_invalid():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    with pytest.raises(TypeError):
        sexp.rclass = rinterface.StrSexpVector(123)


def test__sexp__wrongtypeof():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    cobj = sexp.__sexp__
    sexp = rinterface.StrSexpVector(['a', 'b'])
    assert len(sexp) == 2
    with pytest.raises(ValueError):
        sexp.__sexp__ = cobj


def test__sexp__set():
    x = rinterface.IntSexpVector([1, 2, 3])
    x_s = x.__sexp__
    x_rid = x.rid
    # The Python reference count of the capsule is incremented,
    # not the rpy2 reference count
    assert x.__sexp_refcount__ == 1

    y = rinterface.IntSexpVector([4, 5, 6])
    y_count = y.__sexp_refcount__
    y_rid = y.rid
    assert y_count == 1

    assert x_rid in [elt[0] for elt in rinterface._rinterface.protected_rids()]
    x.__sexp__ = y.__sexp__
    # x_s is still holding a refcount to the capsule
    assert x_rid in [elt[0] for elt in rinterface._rinterface.protected_rids()]
    # when gone, the capsule will be collected and the id no longer preserved
    del(x_s)
    assert x_rid not in [elt[0] for elt in
                         rinterface._rinterface.protected_rids()]

    assert x.rid == y.rid
    assert y_rid == y.rid


@pytest.mark.xfail(reason='WIP')
def test_deepcopy():
    sexp = rinterface.IntSexpVector([1, 2, 3])
    assert sexp.named == 0
    rinterface.baseenv.find("identity")(sexp)
    assert sexp.named >= 2
    sexp2 = sexp.__deepcopy__()
    assert sexp.typeof == sexp2.typeof
    assert list(sexp) == list(sexp2)
    assert not sexp.rsame(sexp2)
    assert sexp2.named == 0
    # should be the same as above, but just in case:
    sexp3 = copy.deepcopy(sexp)
    assert sexp.typeof == sexp3.typeof
    assert list(sexp) == list(sexp3)
    assert not sexp.rsame(sexp3)
    assert sexp3.named == 0


def test_rid():
    globalenv_id = rinterface.baseenv.find('.GlobalEnv').rid
    assert globalenv_id == rinterface.globalenv.rid


def test_NULL_nonzero():
    assert not rinterface.NULL


def test_charsxp_encoding():
    encoding = rinterface.NA_Character.encoding
    assert encoding == rinterface.sexp.CETYPE.CE_NATIVE


def test_charsxp_nchar():
    v = rinterface.StrSexpVector(['abc', 'de', ''])
    cs = v.get_charsxp(0)
    assert cs.nchar() == 3
    cs = v.get_charsxp(1)
    assert cs.nchar() == 2
    cs = v.get_charsxp(2)
    assert cs.nchar() == 0


def test_missingtype():
    assert not rinterface.MissingArg



import pytest
import contextlib
import os
import tempfile

from rpy2.robjects.packages import importr, data
datasets = importr('datasets')
mtcars = data(datasets).fetch('mtcars')['mtcars']
from rpy2.robjects import r

from rpy2.robjects.lib import grdevices


@contextlib.contextmanager
def set_filenames_to_delete():
    todelete = set()
    yield todelete
    for fn in todelete:
        if os.path.exists(fn):
            os.unlink(fn)


@pytest.mark.xfail(os.name == 'nt',
                   reason='Windows produces non-empty file with no plot')
def test_rendertobytes_noplot():
    with grdevices.render_to_bytesio(grdevices.png) as b:
        pass
    assert len(b.getvalue()) == 0


def test_rendertofile():
    fn = tempfile.mktemp(suffix=".png")
    with set_filenames_to_delete() as todelete:
        todelete.add(fn)
        
        with grdevices.render_to_file(grdevices.png,
                                      filename=fn) as d:
            r(''' plot(0) ''')
        assert os.path.exists(fn)


def test_rendertobytes_plot():
    with grdevices.render_to_bytesio(grdevices.png) as b:
        r(''' plot(0) ''')
    assert len(b.getvalue()) > 0



import pytest
import rpy2.rinterface as rinterface

rinterface.initr()


def test_new_invalid():
    x = 1
    with pytest.raises(TypeError):
        rinterface.SexpSymbol(x)


def test_new_missing():
    with pytest.raises(TypeError):
        rinterface.SexpSymbol()


def test_new_fromstring():
    symbol = rinterface.SexpSymbol('pi')
    evalsymbol = rinterface.baseenv['eval'](symbol)
    assert evalsymbol.rid == rinterface.baseenv['pi'].rid


def test_new_str():
    symbol = rinterface.SexpSymbol('pi')
    assert 'pi' == str(symbol)



import array
import pytest
import struct
import sys
import rpy2.rinterface as ri

try:
    import numpy
    has_numpy = True
except ImportError:
    has_numpy = False

ri.initr()


def test_init_from_iter():
    seq = range(3)
    v = ri.IntSexpVector(seq)
    assert len(v) == 3
    for x, y in zip(seq, v):
        assert x == y


def test_init_from_seq_invalid_item():
    seq = (1, 'b', 3)
    with pytest.raises(ValueError):
        ri.IntSexpVector(seq)


@pytest.mark.skip(reason='WIP')
@pytest.mark.skipif(struct.calcsize('P') < 8,
                    reason='Only relevant on 64 architectures.')
def test_init_from_seq_invalid_overflow():
    MAX_INT = ri._rinterface._MAX_INT
    v = ri.IntSexpVector((MAX_INT, 42))
    assert v[0] == MAX_INT
    assert v[1] == 42
    # check 64-bit architecture
    with pytest.raises(OverflowError):
        ri.IntSexpVector((MAX_INT + 1, ))


def _test_from_int_memoryview():
    a = array.array('i', range(3, 103))
    mv = memoryview(a)
    vec = ri.IntSexpVector.from_memoryview(mv)
    assert tuple(range(3, 103)) == tuple(vec)


def test_from_long_memoryview():
    arrtype = 'l'
    if ri._rinterface.ffi.sizeof('int') == ri._rinterface.ffi.sizeof('long'):
        arrtype = 'q'
    a = array.array(arrtype, range(3, 103))
    mv = memoryview(a)
    with pytest.raises(ValueError):
        ri.IntSexpVector.from_memoryview(mv)


def test_from_intarray_object():
    a = array.array('l', range(3, 103))
    vec = ri.IntSexpVector.from_object(a)


def test_from_longarray_object():
    a = array.array('l', range(3, 103))
    vec = ri.IntSexpVector.from_object(a)
    assert tuple(range(3, 103)) == tuple(vec)


def test_from_tuple_object():
    a = tuple(range(3, 103))
    vec = ri.IntSexpVector.from_object(a)
    assert tuple(range(3, 103)) == tuple(vec)


def test_getitem():
    vec = ri.IntSexpVector(range(1, 10))
    assert vec[1] == 2


def test_setitem():
    vec = ri.IntSexpVector(range(1, 10))
    vec[1] = 100
    assert vec[1] == 100


def test_getslice():
    vec = ri.IntSexpVector([1, 2, 3])
    vec = vec[0:2]
    assert len(vec) == 2
    assert vec[0] == 1
    assert vec[1] == 2


def test_getslice_negative():
    vec = ri.IntSexpVector([1, 2, 3])
    vec_s = vec[-2:-1]
    assert len(vec_s) == 1
    assert vec_s[0] == 2


def test_setslice():
    vec = ri.IntSexpVector([1, 2, 3])
    vec[0:2] = ri.IntSexpVector([11, 12])
    assert len(vec) == 3
    assert vec[0] == 11
    assert vec[1] == 12


def test_setslice_negative():
    vec = ri.IntSexpVector([1, 2, 3])
    vec[-2:-1] = ri.IntSexpVector([33, ])
    assert len(vec) == 3
    assert vec[1] == 33


def test_index():
    x = ri.IntSexpVector((1, 2, 3))
    assert x.index(1) == 0
    assert x.index(3) == 2
    with pytest.raises(ValueError):
        x.index(33)
    with pytest.raises(ValueError):
        x.index('a')


def test_getitem_negative_outffbound():
    x = ri.IntSexpVector((1, 2, 3))
    with pytest.raises(IndexError):
        x.__getitem__(-100)


def test_getitem_outofbound():
    x = ri.IntSexpVector((1, 2, 3))
    with pytest.raises(IndexError):
        x.__getitem__(10)


def test_getitem_outofbound_overmaxsize():
    x = ri.IntSexpVector((1, 2, 3))
    with pytest.raises(IndexError):
        x.__getitem__(sys.maxsize + 1)


def test_getslice_missingboundary():
    vec = ri.IntSexpVector(range(1, 11))
    vec_slice = vec[:2]
    assert len(vec_slice) == 2
    assert vec_slice[0] == 1
    assert vec_slice[1] == 2
    vec_slice = vec[8:]
    assert len(vec_slice) == 2
    assert vec_slice[0] == 9
    assert vec_slice[1] == 10
    vec_slice = vec[-2:]
    assert len(vec_slice) == 2
    assert vec_slice[0] == 9
    assert vec_slice[1] == 10


def test_setitem_outffbound():
    vec = ri.IntSexpVector(range(5))
    with pytest.raises(IndexError):
        vec.__setitem__(10, 6)


@pytest.mark.skipif(not has_numpy,
                    reason='numpy currently required for memoryview to work.')
def test_memoryview_2d():
    shape = (5, 2)
    values = tuple(range(10))
    # R arrays are column-major, therefore
    # a slice like [:, :] should look
    # like:
    #
    #  0  5
    #  1  6
    #  2  7
    #  3  8
    #  4  9

    rarray = ri.baseenv['array'](
        ri.IntSexpVector(values),
        dim=ri.IntSexpVector(shape))
    mv = rarray.memoryview()
    assert mv.f_contiguous is True
    assert mv.shape == shape
    assert mv.tolist() == [[0, 5],
                           [1, 6],
                           [2, 7],
                           [3, 8],
                           [4, 9]]
    rarray[0] = 10
    assert mv.tolist() == [[10, 5],
                           [1, 6],
                           [2, 7],
                           [3, 8],
                           [4, 9]]


@pytest.mark.skipif(not has_numpy,
                    reason='numpy currently required for memoryview to work.')
def test_memoryview_3d():
    shape = (5, 2, 3)
    values = tuple(range(30))
    # R arrays are column-major, therefore
    # a slice through the first index in the third dimension should look
    # like:
    #
    #  0  5
    #  1  6
    #  2  7
    #  3  8
    #  4  9

    rarray = ri.baseenv['array'](
        ri.IntSexpVector(values),
        dim=ri.IntSexpVector(shape))
    mv = rarray.memoryview()
    assert mv.f_contiguous is True
    assert mv.shape == shape
    assert tuple(x[0][0] for x in mv.tolist()) == (0, 1, 2, 3, 4)


def test_array_protocol():
    v = ri.IntSexpVector(range(10))
    ai = v.__array_interface__
    assert ai['version'] == 3



import array
import pytest
import rpy2.rinterface as ri

ri.initr()


def test_init_from_bytes_in_seq():
    seq = (b'a', b'b', b'c')
    v = ri.ByteSexpVector(seq)
    assert len(v) == 3
    for x, y in zip(seq, v):
        assert ord(x) == y


def test_init_from_seq_of_bytes():
    seq = (b'a', b'b', b'c')
    v = ri.ByteSexpVector(seq)
    assert len(v) == 3
    for x, y in zip(seq, v):
        assert ord(x) == y


def test_init_from_bytes():
    seq = b'abc'
    v = ri.ByteSexpVector(seq)
    assert len(v) == 3
    for x, y in zip(seq, v):
        assert x == y


def test_init_from_seq_invalid_byte():
    seq = (b'a', [], b'c')
    with pytest.raises(ValueError):
        ri.ByteSexpVector(seq)


def test_from_memoryview():
    a = array.array('b', b'abcdefg')
    mv = memoryview(a)
    vec = ri.ByteSexpVector.from_memoryview(mv)
    assert tuple(b'abcdefg') == tuple(vec)


def test_getitem():
    vec = ri.ByteSexpVector((b'a', b'b', b'c'))
    assert vec[1] == ord(b'b')


def test_getitem_slice():
    vec = ri.ByteSexpVector((b'a', b'b', b'c'))
    assert tuple(vec[:2]) == (ord(b'a'), ord(b'b'))


def test_setitem():
    vec = ri.ByteSexpVector((b'a', b'b', b'c'))
    vec[1] = b'z'
    assert vec[1] == ord(b'z')


def test_setitem_int():
    vec = ri.ByteSexpVector((b'a', b'b', b'c'))
    vec[1] = ord(b'z')
    assert vec[1] == ord(b'z')


def test_setitem_int_invalid():
    vec = ri.ByteSexpVector((b'a', b'b', b'c'))
    with pytest.raises(ValueError):
        vec[1] = 333


def test_setitem_slice():
    values = (b'a', b'b', b'c')
    vec = ri.ByteSexpVector(values)
    vec[:2] = b'yz'
    assert tuple(vec) == tuple(b'yzc')


def test_setitem_slice_invalid():
    values = (b'a', b'b', b'c')
    vec = ri.ByteSexpVector(values)
    with pytest.raises(TypeError):
        vec['foo'] = (333, ord(b'z'))

    with pytest.raises(ValueError):
        vec[:2] = (333, ord(b'z'))



import os
import typing
import warnings
from types import ModuleType
from warnings import warn
import rpy2.rinterface as rinterface
from . import conversion
from rpy2.robjects.functions import (SignatureTranslatedFunction,
                                     docstring_property,
                                     DocumentedSTFunction)
from rpy2.robjects import Environment
from rpy2.robjects.packages_utils import (
                                          default_symbol_r2python,
                                          default_symbol_resolve,
                                          _map_symbols,
                                          _fix_map_symbols
)
import rpy2.robjects.help as rhelp

_require = rinterface.baseenv['require']
_library = rinterface.baseenv['library']
_as_env = rinterface.baseenv['as.environment']
_package_has_namespace = rinterface.baseenv['packageHasNamespace']
_system_file = rinterface.baseenv['system.file']
_get_namespace = rinterface.baseenv['getNamespace']
_get_namespace_version = rinterface.baseenv['getNamespaceVersion']
_get_namespace_exports = rinterface.baseenv['getNamespaceExports']
_loaded_namespaces = rinterface.baseenv['loadedNamespaces']
_globalenv = rinterface.globalenv
_new_env = rinterface.baseenv["new.env"]

StrSexpVector = rinterface.StrSexpVector
# Fetching symbols in the namespace "utils" assumes that "utils" is loaded
# (currently the case by default in R).
_data = rinterface.baseenv['::'](StrSexpVector(('utils', )),
                                 StrSexpVector(('data', )))

_reval = rinterface.baseenv['eval']
_options = rinterface.baseenv['options']


def no_warnings(func):
    """ Decorator to run R functions without warning. """
    def run_withoutwarnings(*args, **kwargs):
        warn_i = _options().do_slot('names').index('warn')
        oldwarn = _options()[warn_i][0]
        _options(warn=-1)
        try:
            res = func(*args, **kwargs)
        except Exception as e:
            # restore the old warn setting before propagating
            # the exception up
            _options(warn=oldwarn)
            raise e
        _options(warn=oldwarn)
        return res
    return run_withoutwarnings


@no_warnings
def _eval_quiet(expr):
    return _reval(expr)


# FIXME: should this be part of the API for rinterface ?
#        (may be it is already the case and there is code
#        duplicaton ?)
def reval(string: str,
          envir: typing.Optional[rinterface.SexpEnvironment] = None):
    """ Evaluate a string as R code
    :param string: R code
    :type string: a :class:`str`
    :param envir: Optional environment to evaluate the R code.
    """
    p = rinterface.parse(string)
    res = _reval(p, envir=envir)
    return res


def quiet_require(name: str, lib_loc=None):
    """ Load an R package /quietly/ (suppressing messages to the console). """
    if lib_loc is None:
        lib_loc = "NULL"
    else:
        lib_loc = "\"%s\"" % (lib_loc.replace('"', '\\"'))
    expr_txt = ("suppressPackageStartupMessages("
                "base::require(%s, lib.loc=%s))"
                % (name, lib_loc))
    expr = rinterface.parse(expr_txt)
    ok = _eval_quiet(expr)
    return ok


class PackageData(object):
    """ Datasets in an R package.
    In R datasets can be distributed with a package.

    Datasets can be:

    - serialized R objects

    - R code (that produces the dataset)

    For a given R packages, datasets are stored separately from the rest
    of the code and are evaluated/loaded lazily.

    The lazy aspect has been conserved and the dataset are only loaded
    or generated when called through the method 'fetch()'.
    """
    _packagename = None
    _lib_loc = None
    _datasets = None

    def __init__(self, packagename, lib_loc=rinterface.NULL):
        self._packagename = packagename
        self._lib_loc

    def _init_setlist(self):
        _datasets = dict()
        # 2D array of information about datatsets
        tmp_m = _data(**{'package': StrSexpVector((self._packagename, )),
                         'lib.loc': self._lib_loc})[2]
        nrows, ncols = tmp_m.do_slot('dim')
        c_i = 2
        for r_i in range(nrows):
            _datasets[tmp_m[r_i + c_i * nrows]] = None
            # FIXME: check if instance methods are overriden
        self._datasets = _datasets

    def names(self):
        """ Names of the datasets"""
        if self._datasets is None:
            self._init_setlist()
        return self._datasets.keys()

    def fetch(self, name):
        """ Fetch the dataset (loads it or evaluates the R associated
        with it.

        In R, datasets are loaded into the global environment by default
        but this function returns an environment that contains the dataset(s).
        """
        if self._datasets is None:
            self._init_setlist()

        if name not in self._datasets:
            raise KeyError('Data set "%s" cannot be found' % name)
        env = _new_env()
        _data(StrSexpVector((name, )),
              **{'package': StrSexpVector((self._packagename, )),
                 'lib.loc': self._lib_loc,
                 'envir': env})
        return Environment(env)


class Package(ModuleType):
    """ Models an R package
    (and can do so from an arbitrary environment - with the caution
    that locked environments should mostly be considered).
     """

    _env = None
    __rname__ = None
    _translation = None
    _rpy2r = None
    _exported_names = None
    _symbol_r2python = None
    __version__ = None
    __rdata__ = None

    def __init__(self, env, name, translation={},
                 exported_names=None, on_conflict='fail',
                 version=None,
                 symbol_r2python=default_symbol_r2python,
                 symbol_resolve=default_symbol_resolve):
        """ Create a Python module-like object from an R environment,
        using the specified translation if defined.

        - env: R environment
        - name: package name
        - translation: `dict` with R names as keys and corresponding Python
                       names as values
        - exported_names: `set` of names/symbols to expose to instance users
        - on_conflict: 'fail' or 'warn' (default: 'fail')
        - version: version string for the package
        - symbol_r2python: function to convert R symbols into Python symbols.
                           The default translate `.` into `_`.
        - symbol_resolve: function to check the Python symbols obtained
                          from `symbol_r2python`.
        """

        super(Package, self).__init__(name)
        self._env = env
        self.__rname__ = name
        self._translation = translation
        mynames = tuple(self.__dict__)
        self._rpy2r = {}
        if exported_names is None:
            exported_names = set(self._env.keys())
        self._exported_names = exported_names
        self._symbol_r2python = symbol_r2python
        self._symbol_resolve = symbol_resolve
        self.__fill_rpy2r__(on_conflict=on_conflict)
        self._exported_names = self._exported_names.difference(mynames)
        self.__version__ = version

    def __update_dict__(self, on_conflict='fail'):
        """ Update the __dict__ according to what is in the R environment """
        for elt in self._rpy2r:
            del(self.__dict__[elt])
        self._rpy2r.clear()
        self.__fill_rpy2r__(on_conflict=on_conflict)

    def __fill_rpy2r__(self, on_conflict='fail'):
        """ Fill the attribute _rpy2r.

        - on_conflict: 'fail' or 'warn' (default: 'fail')
        """
        assert(on_conflict in ('fail', 'warn'))

        name = self.__rname__

        (symbol_mapping,
         conflicts,
         resolutions) = _map_symbols(
             self._env,
             translation=self._translation,
             symbol_r2python=self._symbol_r2python,
             symbol_resolve=self._symbol_resolve
         )
        msg_prefix = ('Conflict when converting R symbols'
                      ' in the package "%s"'
                      ' to Python symbols: \n-' % self.__rname__)
        exception = LibraryError
        _fix_map_symbols(symbol_mapping,
                         conflicts,
                         on_conflict,
                         msg_prefix,
                         exception)
        symbol_mapping.update(resolutions)
        reserved_pynames = set(dir(self))
        for rpyname, rnames in symbol_mapping.items():
            # last paranoid check
            if len(rnames) > 1:
                raise ValueError(
                    'Only one R name should be associated with %s '
                    '(and we have %s)' % (rpyname, str(rnames))
                )
            rname = rnames[0]
            if rpyname in reserved_pynames:
                raise LibraryError('The symbol ' + rname +
                                   ' in the package "' + name + '"' +
                                   ' is conflicting with' +
                                   ' a Python object attribute')
            self._rpy2r[rpyname] = rname
            if (rpyname != rname) and (rname in self._exported_names):
                self._exported_names.remove(rname)
                self._exported_names.add(rpyname)
            try:
                riobj = self._env[rname]
            except rinterface.embedded.RRuntimeError as rre:
                warn(str(rre))
            rpyobj = conversion.rpy2py(riobj)
            if hasattr(rpyobj, '__rname__'):
                rpyobj.__rname__ = rname
            # TODO: shouldn't the original R name be also in the __dict__ ?
            self.__dict__[rpyname] = rpyobj

    def __repr__(self):
        s = super(Package, self).__repr__()
        return 'rpy2.robjects.packages.Package as a %s' % s


# alias
STF = SignatureTranslatedFunction


class SignatureTranslatedPackage(Package):
    """ R package in which the R functions had their signatures
    'translated' (that this the named parameters were made to
    to conform Python's rules for vaiable names)."""
    def __fill_rpy2r__(self, on_conflict='fail'):
        (super(SignatureTranslatedPackage, self)
         .__fill_rpy2r__(on_conflict=on_conflict))
        for name, robj in self.__dict__.items():
            if isinstance(robj, rinterface.Sexp) and \
               robj.typeof == rinterface.RTYPES.CLOSXP:
                self.__dict__[name] = STF(
                    self.__dict__[name],
                    on_conflict=on_conflict,
                    symbol_r2python=self._symbol_r2python,
                    symbol_resolve=self._symbol_resolve
                )


# alias
STP = SignatureTranslatedPackage


class SignatureTranslatedAnonymousPackage(SignatureTranslatedPackage):
    def __init__(self, string, name):
        env = Environment()
        reval(string, env)
        super(SignatureTranslatedAnonymousPackage, self).__init__(env,
                                                                  name)


# alias
STAP = SignatureTranslatedAnonymousPackage


class InstalledSTPackage(SignatureTranslatedPackage):
    @docstring_property(__doc__)
    def __doc__(self):
        doc = list(['Python representation of an R package.'])
        if not self.__rname__:
            doc.append('<No information available>')
        else:
            try:
                doc.append(rhelp.docstring(self.__rname__,
                                           self.__rname__ + '-package',
                                           sections=['\\description']))
            except rhelp.HelpNotFoundError:
                doc.append('[R help was not found]')
        return os.linesep.join(doc)

    def __fill_rpy2r__(self, on_conflict='fail'):
        (super(SignatureTranslatedPackage, self)
         .__fill_rpy2r__(on_conflict=on_conflict))
        for name, robj in self.__dict__.items():
            if isinstance(robj, rinterface.Sexp) and \
               robj.typeof == rinterface.RTYPES.CLOSXP:
                self.__dict__[name] = DocumentedSTFunction(
                    self.__dict__[name],
                    packagename=self.__rname__
                )


class InstalledPackage(Package):
    @docstring_property(__doc__)
    def __doc__(self):
        doc = list(['Python representation of an R package.',
                    'R arguments:', ''])
        if not self.__rname__:
            doc.append('<No information available>')
        else:
            try:
                doc.append(rhelp.docstring(self.__rname__,
                                           self.__rname__ + '-package',
                                           sections=['\\description']))
            except rhelp.HelpNotFoundError:
                doc.append('[R help was not found]')
        return os.linesep.join(doc)


class WeakPackage(Package):
    """
    'Weak' R package, with which looking for symbols results in
    a warning (and a None returned) whenever the desired symbol is
    not found (rather than a traditional `AttributeError`).
    """

    def __getattr__(self, name):
        res = self.__dict__.get(name)
        if res is None:
            warnings.warn(
                "The symbol '%s' is not in this R namespace/package." % name
            )
        return res


class LibraryError(ImportError):
    """ Error occuring when importing an R library """
    pass


class PackageNotInstalledError(LibraryError):
    """ Error occuring because the R package to import is not installed."""
    pass


class InstalledPackages(object):
    """ R packages installed. """
    def __init__(self, lib_loc=None):
        libraryiqr = _library(**{'lib.loc': lib_loc})
        lib_results_i = libraryiqr.do_slot('names').index('results')
        self.lib_results = libraryiqr[lib_results_i]
        self.nrows, self.ncols = self.lib_results.do_slot('dim')
        self.colnames = self.lib_results.do_slot('dimnames')[1]  # column names
        self.lib_packname_i = self.colnames.index('Package')

    def isinstalled(self, packagename: str):
        if not isinstance(packagename, rinterface.StrSexpVector):
            rinterface.StrSexpVector((packagename, ))
        else:
            if len(packagename) > 1:
                raise ValueError("Only specify one package name at a time.")
        nrows = self.nrows
        lib_results, lib_packname_i = self.lib_results, self.lib_packname_i
        for i in range(0+lib_packname_i*nrows,
                       nrows*(lib_packname_i+1),
                       1):
            if lib_results[i] == packagename:
                return True
        return False

    def __iter__(self):
        """ Iterate through rows, yield tuples at each iteration """
        lib_results = self.lib_results
        nrows, ncols = self.nrows, self.ncols
        colrg = range(0, ncols)
        for row_i in range(nrows):
            yield tuple(lib_results[x*nrows+row_i] for x in colrg)


def isinstalled(name: str,
                lib_loc=None):
    """
    Find whether an R package is installed
    :param name: name of an R package
    :param lib_loc: specific location for the R library (default: None)

    :rtype: a :class:`bool`
    """

    instapack = InstalledPackages(lib_loc)
    return instapack.isinstalled(name)


def importr(name: str,
            lib_loc=None,
            robject_translations={},
            signature_translation=True,
            suppress_messages=True,
            on_conflict='fail',
            symbol_r2python=default_symbol_r2python,
            symbol_resolve=default_symbol_resolve,
            data=True):
    """ Import an R package.

    Arguments:

    - name: name of the R package

    - lib_loc: specific location for the R library (default: None)

    - robject_translations: dict (default: {})

    - signature_translation: (True or False)

    - suppress_message: Suppress messages R usually writes on the console
      (defaut: True)

    - on_conflict: 'fail' or 'warn' (default: 'fail')

    - symbol_r2python: function to translate R symbols into Python symbols

    - symbol_resolve: function to check the Python symbol obtained
                      from `symbol_r2python`.

    - data: embed a PackageData objects under the attribute
      name __rdata__ (default: True)

    Return:

    - an instance of class SignatureTranslatedPackage, or of class Package

    """

    if not isinstalled(name, lib_loc=lib_loc):
        raise PackageNotInstalledError(
            'The R package "%s" is not installed.' % name
        )

    if suppress_messages:
        ok = quiet_require(name, lib_loc=lib_loc)
    else:
        ok = _require(name,
                      **{'lib.loc': rinterface.StrSexpVector((lib_loc, ))})[0]
    if not ok:
        raise LibraryError("The R package %s could not be imported" % name)
    if _package_has_namespace(name,
                              _system_file(package=name)):
        env = _get_namespace(name)
        version = _get_namespace_version(name)[0]
        exported_names = set(_get_namespace_exports(name))
    else:
        env = _as_env(rinterface.StrSexpVector(['package:'+name, ]))
        exported_names = None
        version = None

    if signature_translation:
        pack = InstalledSTPackage(env, name,
                                  translation=robject_translations,
                                  exported_names=exported_names,
                                  on_conflict=on_conflict,
                                  version=version,
                                  symbol_r2python=symbol_r2python,
                                  symbol_resolve=symbol_resolve)
    else:
        pack = InstalledPackage(env, name, translation=robject_translations,
                                exported_names=exported_names,
                                on_conflict=on_conflict,
                                version=version,
                                symbol_r2python=symbol_r2python,
                                symbol_resolve=symbol_resolve)
    if data:
        if pack.__rdata__ is not None:
            warn('While importing the R package "%s", the rpy2 Package object '
                 'is masking a translated R symbol "__rdata__" already present'
                 % name)
        pack.__rdata__ = PackageData(name, lib_loc=lib_loc)

    return pack


def data(package):
    """ Return the PackageData for the given package."""
    return package.__rdata__


def wherefrom(symbol: str,
              startenv: rinterface.SexpEnvironment = rinterface.globalenv):
    """ For a given symbol, return the environment
    this symbol is first found in, starting from 'startenv'.
    """
    env = startenv
    while True:
        if symbol in env:
            break
        env = env.enclos
        if env.rsame(rinterface.emptyenv):
            break
    return conversion.rpy2py(env)


class ParsedCode(rinterface.ExprSexpVector):
    pass


class SourceCode(str):

    _parsed = None

    def parse(self):
        if self._parsed is None:
            self._parsed = ParsedCode(rinterface.parse(self))
        return self._parsed

    def as_namespace(self, name):
        """ Name for the namespace """
        return SignatureTranslatedAnonymousPackage(self,
                                                   name)



import pytest

import rpy2.robjects as robjects
rinterface = robjects.rinterface
import array

identical = rinterface.baseenv['identical']
Function = robjects.functions.Function
SignatureTranslatedFunction = robjects.functions.SignatureTranslatedFunction


def test_init_invalid():
    with pytest.raises(ValueError):
        SignatureTranslatedFunction('a')


def test_init():
    ri_f = rinterface.baseenv.find('rank')
    ro_f = SignatureTranslatedFunction(ri_f)        
    assert identical(ri_f, ro_f)[0] is True


def test_init_with_translation():
    ri_f = rinterface.baseenv.find('rank')
    ro_f = SignatureTranslatedFunction(
        ri_f,
        init_prm_translate = {'foo_bar': 'na.last'})
    assert identical(ri_f, ro_f)[0] is True


def test_call():
    ri_f = rinterface.baseenv.find('sum')
    ro_f = robjects.Function(ri_f)

    ro_v = robjects.IntVector(array.array('i', [1,2,3]))

    s = ro_f(ro_v)
    assert s[0] == 6



import pytest
from rpy2.robjects import packages

has_dbplyr = None
try:
    from rpy2.robjects.lib import dbplyr
    has_dbplyr = True
    msg = ''
except packages.PackageNotInstalledError as error:
    has_dbplyr = False
    msg = str(error)


@pytest.mark.skipif(not has_dbplyr,
                    reason=msg)
class TestDplyr(object):

    def test_sql(self):
        sql = dbplyr.sql('count(*)')
        # FIXME: no testing much at the moment...
        assert 'sql' in sql.rclass




import pytest
from rpy2 import rinterface
from rpy2.rinterface import embedded


@pytest.mark.skipif(embedded.rpy2_embeddedR_isinitialized,
                    reason='This test should be run independently of other tests.')
def test_endr():
    rinterface.initr()
    embedded.endr(1)
    assert (embedded.rpy2_embeddedR_isinitialized &
            embedded.RPY_R_Status.ENDED.value) == embedded.RPY_R_Status.ENDED.value




import pytest
import rpy2.rinterface as ri

ri.initr()


def test_init_from_seqr():
    seq = [1+2j, 5+7j, 0+1j]
    v = ri.ComplexSexpVector(seq)
    assert len(v) == 3
    for x, y in zip(seq, v):
        assert x == y


def test_init_from_seq_invalid_item():
    seq = [1+2j, 'a', 0+1j]
    with pytest.raises(TypeError):
        ri.ComplexSexpVector(seq)


def test_getitem():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    assert vec[1] == 5+7j


def test_setitem():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    vec[1] = 100+3j
    assert vec[1] == 100+3j


def test_getslice():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    vec_s = vec[0:2]
    assert len(vec_s) == 2
    assert vec_s[0] == 1+2j
    assert vec_s[1] == 5+7j


def test_getslice_negative():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    vec_s = vec[-2:-1]
    assert len(vec_s) == 1
    assert vec_s[0] == 5+7j


def test_setslice():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    vec[0:2] = ri.ComplexSexpVector([100+3j, 5-5j])
    assert len(vec) == 3
    assert vec[0] == 100+3j
    assert vec[1] == 5-5j


def test_setslice_negative():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    vec[-2:-1] = ri.ComplexSexpVector([100+3j, ])
    assert len(vec) == 3
    assert vec[1] == 100+3j


def test_index():
    vec = ri.ComplexSexpVector([1+2j, 5+7j, 0+1j])
    assert vec.index(5+7j) == 1
    assert vec.index(0+1j) == 2
    with pytest.raises(ValueError):
        vec.index(2)
    with pytest.raises(ValueError):
        vec.index('a')



import pytest
import rpy2.rinterface as rinterface

try:
    import numpy
    has_numpy = True
except ImportError:
    has_numpy = False

rinterface.initr()


def floatEqual(x, y, epsilon=0.00000001):
    return abs(x - y) < epsilon


@pytest.mark.skipif(not has_numpy, reason='Package numpy is not installed.')
def test_array_struct_int():
    px = [1, -2, 3]
    x = rinterface.IntSexpVector(px)
    nx = numpy.asarray(x.memoryview())
    assert nx.dtype.kind == 'i'
    for orig, new in zip(px, nx):
        assert orig == new

    # change value in the Python array... makes it change in the R vector
    nx[1] = 12
    assert x[1] == 12


@pytest.mark.skipif(not has_numpy, reason='Package numpy is not installed.')
def test_array_struct_double():
    px = [1.0, -2.0, 3.0]
    x = rinterface.FloatSexpVector(px)
    nx = numpy.asarray(x.memoryview())
    assert nx.dtype.kind == 'f'
    for orig, new in zip(px, nx):
        assert orig == new

    # change value in the Python array... makes it change in the R vector
    nx[1] = 333.2
    assert x[1] == 333.2


@pytest.mark.skip(reason='WIP')
@pytest.mark.skipif(not has_numpy, reason='Package numpy is not installed.')
def test_array_struct_complex():
    px = [1+2j, 2+5j, -1+0j]
    x = rinterface.ComplexSexpVector(px)
    nx = numpy.asarray(x.memoryview())
    assert nx.dtype.kind == 'c'
    for orig, new in zip(px, nx):
        assert orig == new


@pytest.mark.skipif(not has_numpy, reason='Package numpy is not installed.')
def test_array_struct_boolean():
    px = [True, False, True]
    x = rinterface.BoolSexpVector(px)
    nx = numpy.asarray(x.memoryview())
    # not 'b' as R boolean vectors are array of ints.
    assert nx.dtype.kind == 'i'
    for orig, new in zip(px, nx):
        assert orig == new


@pytest.mark.skipif(not has_numpy, reason='Package numpy is not installed.')
def test_array_shape_len3():
    extract = rinterface.baseenv['[']
    rarray = rinterface.baseenv['array'](
        rinterface.IntSexpVector(range(30)),
        dim=rinterface.IntSexpVector([5, 2, 3]))
    npyarray = numpy.array(rarray.memoryview())
    for i in range(5):
        for j in range(2):
            for k in range(3):
                assert extract(rarray, i+1, j+1, k+1)[0] == npyarray[i, j, k]
